from rdflib.graph import ConjunctiveGraph as ConjunctiveGraph
from rdflib.graph import Graph as Graph
from rdflib.namespace import CSVW as CSVW
from rdflib.namespace import DC as DC
from rdflib.namespace import DCAT as DCAT
from rdflib.namespace import DCTERMS as DCTERMS
from rdflib.namespace import DOAP as DOAP
from rdflib.namespace import FOAF as FOAF
from rdflib.namespace import ODRL2 as ODRL2
from rdflib.namespace import ORG as ORG
from rdflib.namespace import OWL as OWL
from rdflib.namespace import PROF as PROF
from rdflib.namespace import PROV as PROV
from rdflib.namespace import QB as QB
from rdflib.namespace import RDF as RDF
from rdflib.namespace import RDFS as RDFS
from rdflib.namespace import SDO as SDO
from rdflib.namespace import SH as SH
from rdflib.namespace import SKOS as SKOS
from rdflib.namespace import SOSA as SOSA
from rdflib.namespace import SSN as SSN
from rdflib.namespace import TIME as TIME
from rdflib.namespace import VOID as VOID
from rdflib.namespace import XMLNS as XMLNS
from rdflib.namespace import XSD as XSD
from rdflib.namespace import Namespace as Namespace
from rdflib.term import BNode as BNode
from rdflib.term import Literal as Literal
from rdflib.term import URIRef as URIRef
from rdflib.term import Variable as Variable

__version__: str



from collections.abc import Callable
from typing import Any, Union

class Node: ...

class Identifier(Node, str):
    def __new__(cls, value: Any | str | None) -> "Identifier": ...
    def eq(self, other: Any) -> bool: ...
    def neq(self, other: Any) -> bool: ...

class IdentifiedNode(Identifier): ...

class URIRef(IdentifiedNode):
    def toPython(self) -> str: ...
    def n3(self, namespace_manager: Any | None = ...) -> str: ...
    def defrag(self) -> "URIRef": ...
    def de_skolemize(self) -> "BNode": ...

class Genid(URIRef): ...
class RDFLibGenid(Genid): ...

class BNode(IdentifiedNode):
    def toPython(self) -> str: ...
    def n3(self, namespace_manager: Any | None = ...) -> str: ...
    def skolemize(
        self, authority: Any | None = ..., basepath: Any | None = ...
    ) -> "RDFLibGenid": ...

class Literal(Identifier):
    def normalize(self) -> "Literal": ...
    @property
    def value(self) -> Any: ...
    @property
    def language(self) -> str | None: ...
    @property
    def datatype(self) -> Any: ...
    def eq(self, other: Any) -> bool: ...
    def neq(self, other: Any) -> bool: ...
    def n3(self, namespace_manager: Any | None = ...) -> str: ...
    def toPython(self) -> str: ...

class Variable(Identifier):
    def toPython(self) -> str: ...
    def n3(self, namespace_manager: Any | None = ...) -> str: ...



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class ODRL2(DefinedNamespace):
    action: URIRef
    andSequence: URIRef
    assignee: URIRef
    assigneeOf: URIRef
    assigner: URIRef
    assignerOf: URIRef
    attributedParty: URIRef
    attributingParty: URIRef
    compensatedParty: URIRef
    compensatingParty: URIRef
    conflict: URIRef
    consentedParty: URIRef
    consentingParty: URIRef
    consequence: URIRef
    constraint: URIRef
    contractedParty: URIRef
    contractingParty: URIRef
    dataType: URIRef
    duty: URIRef
    failure: URIRef
    function: URIRef
    hasPolicy: URIRef
    implies: URIRef
    includedIn: URIRef
    informedParty: URIRef
    informingParty: URIRef
    inheritAllowed: URIRef
    inheritFrom: URIRef
    inheritRelation: URIRef
    leftOperand: URIRef
    obligation: URIRef
    operand: URIRef
    operator: URIRef
    output: URIRef
    partOf: URIRef
    payeeParty: URIRef
    permission: URIRef
    profile: URIRef
    prohibition: URIRef
    proximity: URIRef
    refinement: URIRef
    relation: URIRef
    remedy: URIRef
    rightOperand: URIRef
    rightOperandReference: URIRef
    scope: URIRef
    source: URIRef
    status: URIRef
    target: URIRef
    timedCount: URIRef
    trackedParty: URIRef
    trackingParty: URIRef
    uid: URIRef
    undefined: URIRef
    unit: URIRef
    xone: URIRef
    All: URIRef
    All2ndConnections: URIRef
    AllConnections: URIRef
    AllGroups: URIRef
    Group: URIRef
    Individual: URIRef
    absolutePosition: URIRef
    absoluteSize: URIRef
    absoluteSpatialPosition: URIRef
    absoluteTemporalPosition: URIRef
    count: URIRef
    dateTime: URIRef
    delayPeriod: URIRef
    deliveryChannel: URIRef
    device: URIRef
    elapsedTime: URIRef
    eq: URIRef
    event: URIRef
    fileFormat: URIRef
    gt: URIRef
    gteq: URIRef
    hasPart: URIRef
    ignore: URIRef
    industry: URIRef
    invalid: URIRef
    isA: URIRef
    isAllOf: URIRef
    isAnyOf: URIRef
    isNoneOf: URIRef
    isPartOf: URIRef
    language: URIRef
    lt: URIRef
    lteq: URIRef
    media: URIRef
    meteredTime: URIRef
    neq: URIRef
    payAmount: URIRef
    percentage: URIRef
    perm: URIRef
    policyUsage: URIRef
    product: URIRef
    prohibit: URIRef
    purpose: URIRef
    recipient: URIRef
    relativePosition: URIRef
    relativeSize: URIRef
    relativeSpatialPosition: URIRef
    relativeTemporalPosition: URIRef
    resolution: URIRef
    spatial: URIRef
    spatialCoordinates: URIRef
    support: URIRef
    system: URIRef
    systemDevice: URIRef
    timeInterval: URIRef
    unitOfCount: URIRef
    version: URIRef
    virtualLocation: URIRef
    Action: URIRef
    Agreement: URIRef
    Assertion: URIRef
    Asset: URIRef
    AssetCollection: URIRef
    AssetScope: URIRef
    ConflictTerm: URIRef
    Constraint: URIRef
    Duty: URIRef
    LeftOperand: URIRef
    LogicalConstraint: URIRef
    Offer: URIRef
    Operator: URIRef
    Party: URIRef
    PartyCollection: URIRef
    PartyScope: URIRef
    Permission: URIRef
    Policy: URIRef
    Privacy: URIRef
    Prohibition: URIRef
    Request: URIRef
    RightOperand: URIRef
    Rule: URIRef
    Set: URIRef
    Ticket: URIRef
    UndefinedTerm: URIRef
    acceptTracking: URIRef
    adHocShare: URIRef
    aggregate: URIRef
    annotate: URIRef
    anonymize: URIRef
    append: URIRef
    appendTo: URIRef
    archive: URIRef
    attachPolicy: URIRef
    attachSource: URIRef
    attribute: URIRef
    commercialize: URIRef
    compensate: URIRef
    concurrentUse: URIRef
    copy: URIRef
    core: URIRef
    delete: URIRef
    derive: URIRef
    digitize: URIRef
    display: URIRef
    distribute: URIRef
    ensureExclusivity: URIRef
    execute: URIRef
    export: URIRef
    extract: URIRef
    extractChar: URIRef
    extractPage: URIRef
    extractWord: URIRef
    give: URIRef
    grantUse: URIRef
    include: URIRef
    index: URIRef
    inform: URIRef
    install: URIRef
    lease: URIRef
    lend: URIRef
    license: URIRef
    modify: URIRef
    move: URIRef
    nextPolicy: URIRef
    obtainConsent: URIRef
    pay: URIRef
    play: URIRef
    present: URIRef
    preview: URIRef
    print: URIRef
    read: URIRef
    reproduce: URIRef
    reviewPolicy: URIRef
    secondaryUse: URIRef
    sell: URIRef
    share: URIRef
    shareAlike: URIRef
    stream: URIRef
    synchronize: URIRef
    textToSpeech: URIRef
    transfer: URIRef
    transform: URIRef
    translate: URIRef
    uninstall: URIRef
    use: URIRef
    watermark: URIRef
    write: URIRef
    writeTo: URIRef



import argparse
from collections.abc import Iterator
from typing import Any, Dict, List, Optional, Tuple

from _typeshed import Incomplete

TRANSLATION_CATALOG: Any
MODULE_NAME: Any
LOGGER: Any
VERSION: Any
PROJECT_URL: str
STANDARD_BAG_INFO_HEADERS: Any
CHECKSUM_ALGOS: Any
DEFAULT_CHECKSUMS: Any
HASH_BLOCK_SIZE: Any
open_text_file: Any
UNICODE_BYTE_ORDER_MARK: str

class Bag:
    valid_files: Any = ...
    valid_directories: Any = ...
    tags: Any = ...
    info: dict[str, str] = ...
    entries: Any = ...
    normalized_filesystem_names: Any = ...
    normalized_manifest_names: Any = ...
    algorithms: Any = ...
    tag_file_name: Any = ...
    path: Any = ...
    def __init__(self, path: Any | None = ...) -> None: ...
    @property
    def algs(self) -> list[str]: ...
    @property
    def version(self) -> str: ...
    def manifest_files(self) -> Iterator[str]: ...
    def tagmanifest_files(self) -> None: ...
    def compare_manifests_with_fs(self) -> tuple[list[str], list[str]]: ...
    def compare_fetch_with_fs(self) -> list[str]: ...
    def payload_files(self) -> Iterator[str]: ...
    def payload_entries(self) -> dict[str, str]: ...
    def save(self, processes: int = ..., manifests: bool = ...) -> None: ...
    def tagfile_entries(self) -> dict[str, str]: ...
    def missing_optional_tagfiles(self) -> Iterator[str]: ...
    def fetch_entries(self) -> Iterator[tuple[str, str, str]]: ...
    def files_to_be_fetched(self) -> Iterator[str]: ...
    def has_oxum(self) -> bool: ...
    def validate(
        self, processes: int = ..., fast: bool = ..., completeness_only: bool = ...
    ) -> bool: ...
    def is_valid(self, fast: bool = ..., completeness_only: bool = ...) -> bool: ...
    def validate_fetch(self) -> None: ...



"""Tests to find local Singularity image."""

import json
import shutil
import subprocess
from collections.abc import Callable
from pathlib import Path
from typing import Any

import pytest
from mypy_extensions import KwArg, VarArg

from cwltool.main import main
from cwltool.singularity import (
    _IMAGES,
    _IMAGES_LOCK,
    _inspect_singularity_sandbox_image,
)

from .util import (
    get_data,
    get_main_output,
    needs_singularity,
    needs_singularity_2_6,
    needs_singularity_3_or_newer,
    working_directory,
)


@pytest.fixture(autouse=True)
def clear_singularity_image_cache() -> None:
    with _IMAGES_LOCK:
        _IMAGES.clear()


@needs_singularity_2_6
def test_singularity_pullfolder(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test singularity respects SINGULARITY_PULLFOLDER."""
    workdir = tmp_path / "working_dir_new"
    workdir.mkdir()
    with working_directory(workdir):
        pullfolder = tmp_path / "pullfolder"
        pullfolder.mkdir()
        with monkeypatch.context() as m:
            result_code, stdout, stderr = get_main_output(
                [
                    "--singularity",
                    get_data("tests/sing_pullfolder_test.cwl"),
                    "--message",
                    "hello",
                ],
                extra_env={"SINGULARITY_PULLFOLDER": str(pullfolder)},
                monkeypatch=m,
            )
            print(stdout)
            print(stderr)
            assert result_code == 0
            image = pullfolder / "debian.img"
            assert image.exists()


@needs_singularity
def test_singularity_workflow(tmp_path: Path) -> None:
    with working_directory(tmp_path):
        error_code, _, stderr = get_main_output(
            [
                "--singularity",
                "--default-container",
                "docker.io/debian:stable-slim",
                "--debug",
                get_data("tests/wf/hello-workflow.cwl"),
                "--usermessage",
                "hello",
            ]
        )
    assert "completed success" in stderr, stderr
    assert error_code == 0


def test_singularity_iwdr(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    singularity_dir = tmp_path / "singularity"
    singularity_dir.mkdir()
    with monkeypatch.context() as m:
        m.setenv("CWL_SINGULARITY_CACHE", str(singularity_dir))
        result_code = main(
            [
                "--singularity",
                "--default-container",
                "docker.io/debian:stable-slim",
                get_data("tests/wf/iwdr-entry.cwl"),
                "--message",
                "hello",
            ]
        )
    singularity_installed = bool(shutil.which("singularity"))
    if singularity_installed:
        assert result_code == 0
    else:
        assert result_code != 0


@needs_singularity
def test_singularity_incorrect_image_pull() -> None:
    result_code, _, stderr = get_main_output(
        [
            "--singularity",
            "--default-container",
            "non-existant-weird-image",
            get_data("tests/wf/hello-workflow.cwl"),
            "--usermessage",
            "hello",
        ]
    )
    assert result_code != 0


@needs_singularity
def test_singularity_local(tmp_path: Path) -> None:
    workdir = tmp_path / "working_dir"
    workdir.mkdir()
    with working_directory(workdir):
        result_code, stdout, stderr = get_main_output(
            [
                "--singularity",
                get_data("tests/sing_pullfolder_test.cwl"),
                "--message",
                "hello",
            ]
        )
        assert result_code == 0


@needs_singularity_2_6
def test_singularity2_docker_image_id_in_tool(tmp_path: Path) -> None:
    workdir = tmp_path / "working_dir"
    workdir.mkdir()
    with working_directory(workdir):
        result_code, stdout, stderr = get_main_output(
            [
                "--singularity",
                get_data("tests/sing_pullfolder_test.cwl"),
                "--message",
                "hello",
            ]
        )
        result_code1, stdout, stderr = get_main_output(
            [
                "--singularity",
                get_data("tests/debian_image_id.cwl"),
                "--message",
                "hello",
            ]
        )
        assert result_code1 == 0


@needs_singularity_3_or_newer
def test_singularity3_docker_image_id_in_tool(tmp_path: Path) -> None:
    workdir = tmp_path / "working_dir"
    workdir.mkdir()
    with working_directory(workdir):
        result_code, stdout, stderr = get_main_output(
            [
                "--singularity",
                get_data("tests/sing_pullfolder_test.cwl"),
                "--message",
                "hello",
            ]
        )
        assert result_code == 0, stderr
        result_code1, stdout, stderr = get_main_output(
            [
                "--singularity",
                get_data("tests/debian_image_id2.cwl"),
                "--message",
                "hello",
            ]
        )
        assert result_code1 == 0


@needs_singularity_3_or_newer
def test_singularity_dockerfile_no_name_no_cache(tmp_path: Path) -> None:
    workdir = tmp_path / "working_dir"
    workdir.mkdir()
    with working_directory(workdir):
        result_code, stdout, stderr = get_main_output(
            [
                "--singularity",
                get_data("tests/sing_dockerfile_test.cwl"),
                "--message",
                "hello",
            ]
        )
        assert result_code == 0, stderr
    assert not (workdir / "bea92b9b6910cbbd2ae602f5bb0f0f27_latest.sif").exists()


@needs_singularity_3_or_newer
def test_singularity_dockerfile_no_name_with_cache(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    workdir = tmp_path / "working_dir"
    workdir.mkdir()
    cachedir = tmp_path / "cache"
    cachedir.mkdir()
    with monkeypatch.context() as m:
        m.setenv("CWL_SINGULARITY_CACHE", str(cachedir))
        with working_directory(workdir):
            result_code, stdout, stderr = get_main_output(
                [
                    "--singularity",
                    get_data("tests/sing_dockerfile_test.cwl"),
                    "--message",
                    "hello",
                ]
            )
            assert result_code == 0, stderr
    assert not (workdir / "bea92b9b6910cbbd2ae602f5bb0f0f27_latest.sif").exists()
    assert (cachedir / "bea92b9b6910cbbd2ae602f5bb0f0f27_latest.sif").exists()


@needs_singularity_3_or_newer
def test_singularity_dockerfile_with_name_no_cache(tmp_path: Path) -> None:
    workdir = tmp_path / "working_dir"
    workdir.mkdir()
    with working_directory(workdir):
        result_code, stdout, stderr = get_main_output(
            [
                "--singularity",
                get_data("tests/sing_dockerfile_named_test.cwl"),
                "--message",
                "hello",
            ]
        )
        assert result_code == 0, stderr
    print(list(workdir.iterdir()))
    assert not (workdir / "bea92b9b6910cbbd2ae602f5bb0f0f27_latest.sif").exists()
    assert not (workdir / "customDebian_latest.sif").exists()


@needs_singularity_3_or_newer
def test_singularity_dockerfile_with_name_with_cache(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    workdir = tmp_path / "working_dir"
    workdir.mkdir()
    cachedir = tmp_path / "cache"
    cachedir.mkdir()
    with monkeypatch.context() as m:
        m.setenv("CWL_SINGULARITY_CACHE", str(cachedir))
        with working_directory(workdir):
            result_code, stdout, stderr = get_main_output(
                [
                    "--singularity",
                    get_data("tests/sing_dockerfile_named_test.cwl"),
                    "--message",
                    "hello",
                ]
            )
            print(list(workdir.iterdir()))
            print(list(cachedir.iterdir()))
            assert result_code == 0, stderr
    assert not (workdir / "bea92b9b6910cbbd2ae602f5bb0f0f27_latest.sif").exists()
    assert not (cachedir / "bea92b9b6910cbbd2ae602f5bb0f0f27_latest.sif").exists()
    assert not (workdir / "customDebian_latest.sif").exists()
    assert (cachedir / "customDebian_latest.sif").exists()


@needs_singularity
def test_singularity_local_sandbox_image(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    workdir = tmp_path / "working_dir"
    workdir.mkdir()
    # build a sandbox image
    container_path = workdir / "container_repo"
    container_path.mkdir()
    cmd = [
        "singularity",
        "build",
        "--sandbox",
        str(container_path / "alpine"),
        "docker://alpine:latest",
    ]
    build = subprocess.run(cmd, capture_output=True, text=True)
    if build.returncode == 0:
        # test that we can work in sub directories
        with working_directory(workdir):
            result_code, _, _ = get_main_output(
                [
                    "--singularity",
                    "--disable-pull",
                    get_data("tests/sing_local_sandbox_test.cwl"),
                    "--message",
                    "hello",
                ]
            )
            assert result_code == 0
            result_code, _, _ = get_main_output(
                [
                    "--singularity",
                    "--disable-pull",
                    get_data("tests/sing_local_sandbox_img_id_test.cwl"),
                    "--message",
                    "hello",
                ]
            )
            assert result_code == 0
        # test with --singularity-sandbox-path option:
        result_code, out, err = get_main_output(
            [
                "--singularity",
                "--disable-pull",
                "--singularity-sandbox-path",
                f"{workdir}",
                get_data("tests/sing_local_sandbox_test.cwl"),
                "--message",
                "hello",
            ]
        )
        assert result_code == 0

        # test with CWL_SINGULARITY_IMAGES env variable set:
        with monkeypatch.context() as m:
            m.setenv("CWL_SINGULARITY_IMAGES", str(workdir))
            result_code, _, _ = get_main_output(
                [
                    "--singularity",
                    "--disable-pull",
                    get_data("tests/sing_local_sandbox_test.cwl"),
                    "--message",
                    "hello",
                ]
            )
            assert result_code == 0
    else:
        pytest.skip(f"Failed to build the singularity image: {build.stderr}")


@needs_singularity
def test_singularity_inspect_image(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test inspect a real image works."""
    workdir = tmp_path / "working_dir"
    workdir.mkdir()
    repo_path = workdir / "container_repo"
    image_path = repo_path / "alpine"

    # test image exists
    repo_path.mkdir()
    cmd = [
        "singularity",
        "build",
        "--sandbox",
        str(image_path),
        "docker://alpine:latest",
    ]
    build = subprocess.run(cmd, capture_output=True, text=True)
    if build.returncode == 0:
        # Verify the path is a correct container image
        res_inspect = _inspect_singularity_sandbox_image(str(image_path))
        assert res_inspect is True
    else:
        pytest.skip(f"singularity sandbox image build didn't worked: {build.stderr}")


class _DummyResult:  # noqa: B903
    def __init__(self, rc: int, out: str) -> None:
        self.returncode = rc
        self.stdout = out


def _make_run_result(
    returncode: int, stdout: str
) -> Callable[[VarArg(Any), KwArg(Any)], _DummyResult]:
    """Mock subprocess.run returning returncode and stdout."""

    def _runner(*args: Any, **kwargs: Any) -> _DummyResult:
        return _DummyResult(returncode, stdout)

    return _runner


def test_json_decode_error_branch(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test json can't decode inspect result."""
    monkeypatch.setattr("cwltool.singularity.run", _make_run_result(0, "not-a-json"))

    def _raise_json_error(s: str) -> None:
        # construct and raise an actual JSONDecodeError
        raise json.JSONDecodeError("Expecting value", s, 0)

    monkeypatch.setattr("json.loads", _raise_json_error)

    assert _inspect_singularity_sandbox_image("/tmp/image") is False


def test_singularity_sandbox_image_not_exists() -> None:
    image_path = "/tmp/not_existing/image"
    res_inspect = _inspect_singularity_sandbox_image(image_path)
    assert res_inspect is False


def test_singularity_sandbox_not_an_image(tmp_path: Path) -> None:
    image_path = tmp_path / "image"
    image_path.mkdir()
    res_inspect = _inspect_singularity_sandbox_image(str(image_path))
    assert res_inspect is False


def test_inspect_image_wrong_sb_call(monkeypatch: pytest.MonkeyPatch) -> None:

    def mock_failed_subprocess(*args: Any, **kwargs: Any) -> None:
        raise subprocess.CalledProcessError(returncode=1, cmd=args[0])

    monkeypatch.setattr("cwltool.singularity.run", mock_failed_subprocess)
    res_inspect = _inspect_singularity_sandbox_image("/tmp/container_repo/alpine")
    assert res_inspect is False



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import importlib.metadata
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import time
from datetime import datetime, timezone

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

build_date = datetime.fromtimestamp(
    int(os.environ.get("SOURCE_DATE_EPOCH", time.time())), timezone.utc
)
project = "Common Workflow Language reference implementation"
copyright = f"2019 — {build_date.year}, Peter Amstutz and contributors to the CWL Project"
author = "Peter Amstutz and Common Workflow Language Project contributors"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.inheritance_diagram",
    "autoapi.extension",
    "sphinx_autodoc_typehints",
    "sphinx_rtd_theme",
    "sphinxcontrib.autoprogram",
]

autosectionlabel_prefix_document = True

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "schema_salad": ("https://schema-salad.readthedocs.io/en/stable/", None),
    "rdflib": ("https://rdflib.readthedocs.io/en/stable/", None),
    "cwl_utils": ("https://cwl-utils.readthedocs.io/en/stable/", None),
    # "ruamel.yaml": ("https://yaml.readthedocs.io/en/stable/", None),
}


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# html_logo = "_static/logo.png"
html_favicon = "_static/favicon.ico"

html_theme_options = {
    "collapse_navigation": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

release = importlib.metadata.version("cwltool")
version = ".".join(release.split(".")[:2])

autoapi_dirs = ["../cwltool"]
autodoc_typehints = "description"
autoapi_keep_files = True
autoapi_ignore = ["*migrations*", "*.pyi"]
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-inheritance-diagram",
    "show-module-summary",
    "imported-members",
    "special-members",
]
# sphinx-autodoc-typehints
always_document_param_types = True
# If False, do not add type info for undocumented parameters.
# If True, add stub documentation for undocumented parameters to be able to add type info.



"""Tests of the experimental MPI extension."""

import json
import os.path
import sys
from collections.abc import Generator, MutableMapping
from importlib.resources import files
from io import StringIO
from pathlib import Path
from typing import Any, cast

import pytest
from ruamel.yaml.comments import CommentedMap, CommentedSeq
from schema_salad.avro.schema import Names
from schema_salad.ref_resolver import file_uri
from schema_salad.utils import yaml_no_ts

import cwltool.load_tool
import cwltool.singularity
import cwltool.udocker
from cwltool.command_line_tool import CommandLineTool
from cwltool.context import RuntimeContext
from cwltool.main import main
from cwltool.mpi import MpiConfig, MPIRequirementName

from .util import get_data, working_directory


def test_mpi_conf_defaults() -> None:
    mpi = MpiConfig()
    assert mpi.runner == "mpirun"
    assert mpi.nproc_flag == "-n"
    assert mpi.default_nproc == 1
    assert mpi.extra_flags == []
    assert mpi.env_pass == []
    assert mpi.env_pass_regex == []
    assert mpi.env_set == {}


def test_mpi_conf_unknownkeys() -> None:
    with pytest.raises(TypeError):
        MpiConfig(runner="mpiexec", foo="bar")  # type: ignore


@pytest.fixture(scope="class")
def fake_mpi_conf(tmp_path_factory: Any) -> Generator[str, None, None]:
    """
    Make a super simple mpirun-alike for applications that don't actually use MPI.

    It just runs the command multiple times (in serial).

    Then create a platform MPI config YAML file that should make it work
    for the testing examples.
    """
    mpirun_text = """#!{interpreter}
import argparse
import sys
import subprocess
from io import StringIO
from typing import List

def make_parser():
    p = argparse.ArgumentParser()
    p.add_argument("--num", type=int, help="number of times to run the application")
    p.add_argument(
        "--no-fail", help="add this flag to actually work", action="store_true"
    )
    p.add_argument(
        "progargs", nargs=argparse.REMAINDER, help="The program and its arguments"
    )
    return p

class Runner:
    def __init__(self):
        if sys.stdin.isatty():
            self.indata = None
        else:
            self.indata = sys.stdin.read().encode(sys.stdin.encoding)

    def run_once(self, args: list[str]):
        subprocess.run(
            args, input=self.indata, stdout=sys.stdout, stderr=sys.stderr
        ).check_returncode()

    def run_many(self, n: int, args: list[str]):
        for i in range(n):
            self.run_once(args)

if __name__ == "__main__":
    args = make_parser().parse_args()
    assert args.no_fail == True, "Didn't set the --no-fail flag"
    r = Runner()
    r.run_many(args.num, args.progargs)
""".format(interpreter=sys.executable)
    mpitmp = tmp_path_factory.mktemp("fake_mpi")
    mpirun_file = mpitmp / "fake_mpirun"
    mpirun_file.write_text(mpirun_text)
    mpirun_file.chmod(0o755)

    plat_conf = {
        "runner": str(mpirun_file),
        "nproc_flag": "--num",
        "extra_flags": ["--no-fail"],
        "env_set": {"TEST_MPI_FOO": "bar"},
        "env_pass": ["USER"],
    }
    plat_conf_file = mpitmp / "plat_mpi.yml"
    yaml = yaml_no_ts()
    yaml.dump(plat_conf, plat_conf_file)

    yield str(plat_conf_file)

    plat_conf_file.unlink()
    mpirun_file.unlink()
    mpitmp.rmdir()


def make_processes_input(np: int, tmp_path: Path) -> Path:
    input_file = tmp_path / "input.yml"
    with input_file.open("w") as f:
        f.write("processes: %d\n" % np)
    return input_file


def cwltool_args(fake_mpi_conf: str) -> list[str]:
    return ["--enable-ext", "--enable-dev", "--mpi-config-file", fake_mpi_conf]


class TestMpiRun:
    def test_fake_mpi_config(self, fake_mpi_conf: str) -> None:
        conf_obj = MpiConfig.load(fake_mpi_conf)
        runner = conf_obj.runner
        assert os.path.dirname(runner) == os.path.dirname(fake_mpi_conf)
        assert os.path.basename(runner) == "fake_mpirun"
        assert conf_obj.nproc_flag == "--num"
        assert conf_obj.default_nproc == 1
        assert conf_obj.extra_flags == ["--no-fail"]

    def test_simple_mpi_tool(self, fake_mpi_conf: str, tmp_path: Path) -> None:
        stdout = StringIO()
        stderr = StringIO()
        with working_directory(tmp_path):
            rc = main(
                argsl=cwltool_args(fake_mpi_conf) + [get_data("tests/wf/mpi_simple.cwl")],
                stdout=stdout,
                stderr=stderr,
            )
            assert rc == 0

            output = json.loads(stdout.getvalue())
            pid_path = output["pids"]["path"]
            with open(pid_path) as pidfile:
                pids = [int(line) for line in pidfile]
            assert len(pids) == 2

    def test_simple_mpi_nproc_expr(self, fake_mpi_conf: str, tmp_path: Path) -> None:
        np = 4
        input_file = make_processes_input(np, tmp_path)
        stdout = StringIO()
        stderr = StringIO()
        with working_directory(tmp_path):
            rc = main(
                argsl=cwltool_args(fake_mpi_conf)
                + [get_data("tests/wf/mpi_expr.cwl"), str(input_file)],
                stdout=stdout,
                stderr=stderr,
            )
            assert rc == 0

            output = json.loads(stdout.getvalue())
            pid_path = output["pids"]["path"]
            with open(pid_path) as pidfile:
                pids = [int(line) for line in pidfile]
            assert len(pids) == np

    def test_mpi_workflow(self, fake_mpi_conf: str, tmp_path: Path) -> None:
        np = 3
        input_file = make_processes_input(np, tmp_path)
        stdout = StringIO()
        stderr = StringIO()
        with working_directory(tmp_path):
            rc = main(
                argsl=cwltool_args(fake_mpi_conf)
                + [get_data("tests/wf/mpi_simple_wf.cwl"), str(input_file)],
                stdout=stdout,
                stderr=stderr,
            )
            assert rc == 0

            output = json.loads(stdout.getvalue())
            lc_path = output["line_count"]["path"]
            with open(lc_path) as lc_file:
                lc = int(lc_file.read())
                assert lc == np

    def test_environment(
        self, fake_mpi_conf: str, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        stdout = StringIO()
        stderr = StringIO()
        with monkeypatch.context() as m:
            m.setenv("USER", "tester")
            with working_directory(tmp_path):
                rc = main(
                    argsl=cwltool_args(fake_mpi_conf) + [get_data("tests/wf/mpi_env.cwl")],
                    stdout=stdout,
                    stderr=stderr,
                )
                assert rc == 0

                output = json.loads(stdout.getvalue())
                env_path = output["environment"]["path"]
                with open(env_path) as envfile:
                    e = {}
                    for line in envfile:
                        k, v = line.strip().split("=", 1)
                        e[k] = v
                assert e["USER"] == "tester"
                assert e["TEST_MPI_FOO"] == "bar"


def test_env_passing(monkeypatch: pytest.MonkeyPatch) -> None:
    """Confirm that MPI extension passes environment variables correctly."""
    config = MpiConfig(
        env_pass=["A", "B", "LONG_NAME"],
        env_pass_regex=["TOOLNAME", "MPI_.*_CONF"],
    )

    env: MutableMapping[str, str] = {}

    with monkeypatch.context() as m:
        m.setattr(os, "environ", {})
        env = {}
        config.pass_through_env_vars(env)
        assert env == {}

    with monkeypatch.context() as m:
        m.setattr(os, "environ", {"A": "a"})
        env = {}
        config.pass_through_env_vars(env)
        assert env == {"A": "a"}

    with monkeypatch.context() as m:
        m.setattr(os, "environ", {"A": "a", "C": "c"})
        env = {}
        config.pass_through_env_vars(env)
        assert env == {"A": "a"}

    with monkeypatch.context() as m:
        m.setattr(os, "environ", {"A": "a", "B": "b", "C": "c"})
        env = {"PATH": "one:two:three", "HOME": "/tmp/dir", "TMPDIR": "/tmp/dir"}
        config.pass_through_env_vars(env)
        assert env == {
            "PATH": "one:two:three",
            "HOME": "/tmp/dir",
            "TMPDIR": "/tmp/dir",
            "A": "a",
            "B": "b",
        }

    with monkeypatch.context() as m:
        m.setattr(os, "environ", {"TOOLNAME": "foobar"})
        env = {}
        config.pass_through_env_vars(env)
        assert env == {"TOOLNAME": "foobar"}

    with monkeypatch.context() as m:
        m.setattr(os, "environ", {"_TOOLNAME_": "foobar"})
        env = {}
        config.pass_through_env_vars(env)
        # Cos we are matching not searching
        assert env == {}

    with monkeypatch.context() as m:
        m.setattr(os, "environ", {"MPI_A_CONF": "A", "MPI_B_CONF": "B"})

        env = {}
        config.pass_through_env_vars(env)
        # Cos we are matching not searching
        assert env == {"MPI_A_CONF": "A", "MPI_B_CONF": "B"}


# Reading the schema is super slow - cache for the session
@pytest.fixture(scope="session")
def schema_ext11() -> Generator[Names, None, None]:
    ext11 = files("cwltool").joinpath("extensions-v1.1.yml").read_text("utf-8")
    cwltool.process.use_custom_schema("v1.1", "http://commonwl.org/cwltool", ext11)
    schema = cwltool.process.get_schema("v1.1")[1]
    assert isinstance(schema, Names)
    yield schema


mpiReq = CommentedMap({"class": MPIRequirementName, "processes": 1})
containerReq = CommentedMap({"class": "DockerRequirement"})
basetool = CommentedMap(
    {
        "cwlVersion": "v1.1",
        "class": "CommandLineTool",
        "inputs": CommentedSeq(),
        "outputs": CommentedSeq(),
    }
)


def mk_tool(
    schema: Names,
    opts: list[str],
    reqs: list[CommentedMap] | None = None,
    hints: list[CommentedMap] | None = None,
) -> tuple[RuntimeContext, CommandLineTool]:
    tool = basetool.copy()

    if reqs is not None:
        tool["requirements"] = CommentedSeq(reqs)
    if hints is not None:
        tool["hints"] = CommentedSeq(hints)

    args = cwltool.argparser.arg_parser().parse_args(opts)
    args.enable_ext = True
    args.basedir = os.path.dirname(os.path.abspath("."))
    rc = RuntimeContext(vars(args))
    lc = cwltool.main.setup_loadingContext(None, rc, args)
    lc.avsc_names = schema
    tool["id"] = file_uri(os.path.abspath("./mktool.cwl"))
    assert lc.loader is not None
    lc.loader.idx[tool["id"]] = tool
    return rc, cast(CommandLineTool, cwltool.load_tool.load_tool(tool, lc))


def test_singularity(schema_ext11: Names) -> None:
    rc, clt = mk_tool(schema_ext11, ["--singularity"], reqs=[mpiReq, containerReq])
    clt._init_job({}, rc)
    jr = clt.make_job_runner(rc)
    assert jr is cwltool.singularity.SingularityCommandLineJob


def test_udocker(schema_ext11: Names) -> None:
    rc, clt = mk_tool(schema_ext11, ["--udocker"], reqs=[mpiReq, containerReq])
    clt._init_job({}, rc)
    jr = clt.make_job_runner(rc)
    assert jr is cwltool.udocker.UDockerCommandLineJob


def test_docker_hint(schema_ext11: Names) -> None:
    # Docker hint, MPI required
    rc, clt = mk_tool(schema_ext11, [], hints=[containerReq], reqs=[mpiReq])
    clt._init_job({}, rc)
    jr = clt.make_job_runner(rc)
    assert jr is cwltool.job.CommandLineJob


def test_docker_required(schema_ext11: Names) -> None:
    # Docker required, MPI hinted
    rc, clt = mk_tool(schema_ext11, [], reqs=[containerReq], hints=[mpiReq])
    clt._init_job({}, rc)
    jr = clt.make_job_runner(rc)
    assert jr is cwltool.docker.DockerCommandLineJob


def test_docker_mpi_both_required(schema_ext11: Names) -> None:
    # Both required - error
    rc, clt = mk_tool(schema_ext11, [], reqs=[mpiReq, containerReq])
    with pytest.raises(cwltool.errors.UnsupportedRequirement):
        clt._init_job({}, rc)
    clt.make_job_runner(rc)


def test_docker_mpi_both_hinted(schema_ext11: Names) -> None:
    # Both hinted - error
    rc, clt = mk_tool(schema_ext11, [], hints=[mpiReq, containerReq])
    with pytest.raises(cwltool.errors.UnsupportedRequirement):
        clt._init_job({}, rc)
    clt.make_job_runner(rc)



from collections.abc import Iterator, Mapping
from typing import IO, Any, Dict, List, Optional, SupportsIndex, Tuple, overload

from rdflib import URIRef, Variable
from rdflib.term import Identifier

class ResultRow(tuple["Identifier", ...]):
    def __new__(
        cls, values: Mapping[Variable, Identifier], labels: list[Variable]
    ) -> ResultRow: ...
    def __getattr__(self, name: str) -> Identifier: ...
    @overload
    def __getitem__(self, name: str) -> Identifier: ...
    @overload
    def __getitem__(self, __x: SupportsIndex) -> Identifier: ...
    @overload
    def __getitem__(self, __x: slice) -> tuple[Identifier, ...]: ...
    def get(self, name: str, default: Any | None = ...) -> Identifier: ...
    def asdict(self) -> dict[str, Identifier]: ...

class Result:
    type: Any
    vars: Any
    askAnswer: Any
    graph: Any
    def __init__(self, type_: str) -> None: ...
    bindings: Any
    def __iter__(self) -> Iterator[bool | ResultRow]: ...
    @staticmethod
    def parse(
        source: IO[Any] | None = ...,
        format: str | None = ...,
        content_type: str | None = ...,
        **kwargs: Any,
    ) -> Result: ...
    def serialize(
        self,
        destination: str | IO[Any] | None = ...,
        encoding: str = ...,
        format: str = ...,
        **args: Any,
    ) -> bytes | None: ...



from pytest import LogCaptureFixture

from .util import get_main_output


def test_content_types(caplog: LogCaptureFixture) -> None:
    for _ in ("js_output.cwl", "js_output_workflow.cwl"):
        commands = [
            "https://raw.githubusercontent.com/common-workflow-language/common-workflow-language/main/v1.0/v1.0/test-cwl-out2.cwl",
            "https://www.cnn.com",
        ]
        error_code, _, stderr = get_main_output(commands)

        found = False
        for record in caplog.records:
            if record.name == "salad" and "got content-type of 'text/html'" in record.message:
                found = True
                break
        assert found
        assert error_code == 1, stderr



from unittest import mock
from unittest.mock import MagicMock

import pytest
from schema_salad.avro import schema

from cwltool.builder import Builder
from cwltool.command_line_tool import CommandLineTool
from cwltool.context import LoadingContext, RuntimeContext
from cwltool.cuda import cuda_version_and_device_count
from cwltool.errors import WorkflowException
from cwltool.job import CommandLineJob
from cwltool.load_tool import load_tool
from cwltool.main import main
from cwltool.process import use_custom_schema
from cwltool.stdfsaccess import StdFsAccess
from cwltool.update import INTERNAL_VERSION
from cwltool.utils import CWLObjectType

from .util import get_data, needs_docker, needs_singularity_3_or_newer

cuda_version = cuda_version_and_device_count()


@needs_docker
@pytest.mark.skipif(cuda_version[0] == "", reason="nvidia-smi required for CUDA not detected")
def test_cuda_docker() -> None:
    params = [
        "--enable-ext",
        get_data("tests/wf/nvidia-smi-container.cwl"),
    ]
    assert main(params) == 0


@needs_singularity_3_or_newer
@pytest.mark.skipif(cuda_version[0] == "", reason="nvidia-smi required for CUDA not detected")
def test_cuda_singularity() -> None:
    params = [
        "--enable-ext",
        "--singularity",
        get_data("tests/wf/nvidia-smi-container.cwl"),
    ]
    assert main(params) == 0


@pytest.mark.skipif(cuda_version[0] == "", reason="nvidia-smi required for CUDA not detected")
def test_cuda_no_container() -> None:
    params = [
        "--enable-ext",
        get_data("tests/wf/nvidia-smi.cwl"),
    ]
    assert main(params) == 0


@pytest.mark.skipif(cuda_version[0] == "", reason="nvidia-smi required for CUDA not detected")
def test_cuda_cc_list() -> None:
    params = [
        "--enable-ext",
        get_data("tests/wf/nvidia-smi-cc.cwl"),
    ]
    assert main(params) == 0


def _makebuilder(cudaReq: CWLObjectType) -> Builder:
    return Builder(
        {},
        [],
        [],
        {},
        schema.Names(),
        [cudaReq],
        [],
        {"cudaDeviceCount": 1},
        None,
        None,
        StdFsAccess,
        StdFsAccess(""),
        None,
        0.1,
        False,
        False,
        False,
        "no_listing",
        "",
        "",
        "",
        INTERNAL_VERSION,
        "docker",
    )


@mock.patch("subprocess.check_output")
@mock.patch("os.makedirs")
def test_cuda_job_setup_check(makedirs: MagicMock, check_output: MagicMock) -> None:
    runtime_context = RuntimeContext({})

    cudaReq: CWLObjectType = {
        "class": "http://commonwl.org/cwltool#CUDARequirement",
        "cudaVersionMin": "1.0",
        "cudaComputeCapability": "1.0",
    }
    builder = _makebuilder(cudaReq)

    check_output.return_value = """
<nvidia>
<attached_gpus>1</attached_gpus>
<cuda_version>1.0</cuda_version>
</nvidia>
"""

    jb = CommandLineJob(builder, {}, CommandLineTool.make_path_mapper, [], [], "")
    jb._setup(runtime_context)


@mock.patch("subprocess.check_output")
@mock.patch("os.makedirs")
def test_cuda_job_setup_check_err(makedirs: MagicMock, check_output: MagicMock) -> None:
    runtime_context = RuntimeContext({})

    cudaReq: CWLObjectType = {
        "class": "http://commonwl.org/cwltool#CUDARequirement",
        "cudaVersionMin": "2.0",
        "cudaComputeCapability": "1.0",
    }
    builder = _makebuilder(cudaReq)

    check_output.return_value = """
<nvidia>
<attached_gpus>1</attached_gpus>
<cuda_version>1.0</cuda_version>
</nvidia>
"""
    jb = CommandLineJob(builder, {}, CommandLineTool.make_path_mapper, [], [], "")
    with pytest.raises(WorkflowException):
        jb._setup(runtime_context)


@mock.patch("subprocess.check_output")
@mock.patch("os.makedirs")
def test_cuda_job_setup_check_err_empty_attached_gpus(
    makedirs: MagicMock, check_output: MagicMock, caplog: pytest.LogCaptureFixture
) -> None:
    runtime_context = RuntimeContext({})

    cudaReq: CWLObjectType = {
        "class": "http://commonwl.org/cwltool#CUDARequirement",
        "cudaVersionMin": "1.0",
        "cudaComputeCapability": "1.0",
    }
    builder = _makebuilder(cudaReq)

    check_output.return_value = """
<nvidia>
<attached_gpus></attached_gpus>
<cuda_version>1.0</cuda_version>
</nvidia>
"""

    jb = CommandLineJob(builder, {}, CommandLineTool.make_path_mapper, [], [], "")
    with pytest.raises(WorkflowException):
        jb._setup(runtime_context)
    assert (
        "Error checking CUDA version with nvidia-smi. Missing 'attached_gpus' or it is empty."
        in caplog.text
    )


@mock.patch("subprocess.check_output")
@mock.patch("os.makedirs")
def test_cuda_job_setup_check_err_empty_missing_attached_gpus(
    makedirs: MagicMock, check_output: MagicMock, caplog: pytest.LogCaptureFixture
) -> None:
    runtime_context = RuntimeContext({})

    cudaReq: CWLObjectType = {
        "class": "http://commonwl.org/cwltool#CUDARequirement",
        "cudaVersionMin": "1.0",
        "cudaComputeCapability": "1.0",
    }
    builder = _makebuilder(cudaReq)

    check_output.return_value = """
<nvidia>
<cuda_version>1.0</cuda_version>
</nvidia>
"""

    jb = CommandLineJob(builder, {}, CommandLineTool.make_path_mapper, [], [], "")
    with pytest.raises(WorkflowException):
        jb._setup(runtime_context)
    assert (
        "Error checking CUDA version with nvidia-smi. Missing 'attached_gpus' or it is empty."
        in caplog.text
    )


@mock.patch("subprocess.check_output")
@mock.patch("os.makedirs")
def test_cuda_job_setup_check_err_empty_cuda_version(
    makedirs: MagicMock, check_output: MagicMock, caplog: pytest.LogCaptureFixture
) -> None:
    runtime_context = RuntimeContext({})

    cudaReq: CWLObjectType = {
        "class": "http://commonwl.org/cwltool#CUDARequirement",
        "cudaVersionMin": "1.0",
        "cudaComputeCapability": "1.0",
    }
    builder = _makebuilder(cudaReq)

    check_output.return_value = """
<nvidia>
<attached_gpus>1</attached_gpus>
<cuda_version></cuda_version>
</nvidia>
"""

    jb = CommandLineJob(builder, {}, CommandLineTool.make_path_mapper, [], [], "")
    with pytest.raises(WorkflowException):
        jb._setup(runtime_context)
    assert (
        "Error checking CUDA version with nvidia-smi. Missing 'cuda_version' or it is empty."
        in caplog.text
    )


@mock.patch("subprocess.check_output")
@mock.patch("os.makedirs")
def test_cuda_job_setup_check_err_missing_cuda_version(
    makedirs: MagicMock, check_output: MagicMock, caplog: pytest.LogCaptureFixture
) -> None:
    runtime_context = RuntimeContext({})

    cudaReq: CWLObjectType = {
        "class": "http://commonwl.org/cwltool#CUDARequirement",
        "cudaVersionMin": "1.0",
        "cudaComputeCapability": "1.0",
    }
    builder = _makebuilder(cudaReq)

    check_output.return_value = """
<nvidia>
<attached_gpus>1</attached_gpus>
</nvidia>
"""

    jb = CommandLineJob(builder, {}, CommandLineTool.make_path_mapper, [], [], "")
    with pytest.raises(WorkflowException):
        jb._setup(runtime_context)
    assert (
        "Error checking CUDA version with nvidia-smi. Missing 'cuda_version' or it is empty."
        in caplog.text
    )


@mock.patch("subprocess.check_output")
@mock.patch("os.makedirs")
def test_cuda_job_setup_check_err_wrong_type_cuda_version(
    makedirs: MagicMock, check_output: MagicMock, caplog: pytest.LogCaptureFixture
) -> None:
    runtime_context = RuntimeContext({})

    cudaReq: CWLObjectType = {
        "class": "http://commonwl.org/cwltool#CUDARequirement",
        "cudaVersionMin": "1.0",
        "cudaComputeCapability": "1.0",
    }
    builder = _makebuilder(cudaReq)

    check_output.return_value = """
<nvidia>
<attached_gpus>1</attached_gpus>
<cuda_version><subelement /></cuda_version>
</nvidia>
"""

    jb = CommandLineJob(builder, {}, CommandLineTool.make_path_mapper, [], [], "")
    with pytest.raises(WorkflowException):
        jb._setup(runtime_context)
    assert (
        "Error checking CUDA version with nvidia-smi. "
        "Either 'attached_gpus' or 'cuda_version' was not a text node" in caplog.text
    )


def test_cuda_eval_resource_range() -> None:
    with open(get_data("extensions-v1.1.yml")) as res:
        use_custom_schema("v1.2", "http://commonwl.org/cwltool", res.read())

    joborder: CWLObjectType = {}
    loadingContext = LoadingContext({"do_update": True})
    runtime_context = RuntimeContext({})

    tool = load_tool(get_data("tests/wf/nvidia-smi-range.cwl"), loadingContext)
    builder = _makebuilder(tool.requirements[0])
    builder.job = joborder

    resources = tool.evalResources(builder, runtime_context)

    assert resources["cudaDeviceCount"] == 2


def test_cuda_eval_resource_max() -> None:
    with open(get_data("extensions-v1.1.yml")) as res:
        use_custom_schema("v1.2", "http://commonwl.org/cwltool", res.read())

    joborder: CWLObjectType = {}
    loadingContext = LoadingContext({"do_update": True})
    runtime_context = RuntimeContext({})

    tool = load_tool(get_data("tests/wf/nvidia-smi-max.cwl"), loadingContext)
    builder = _makebuilder(tool.requirements[0])
    builder.job = joborder

    resources = tool.evalResources(builder, runtime_context)

    assert resources["cudaDeviceCount"] == 4



"""InitialWorkDirRequirement related tests."""

import json
import re
from pathlib import Path
from stat import S_IWGRP, S_IWOTH, S_IWRITE
from typing import Any

import pytest

from cwltool.factory import Factory
from cwltool.main import main

from .util import get_data, get_main_output, needs_docker, needs_singularity


def test_newline_in_entry() -> None:
    """Files in a InitialWorkingDirectory are created with a newline character."""
    factory = Factory()
    echo = factory.make(get_data("tests/wf/iwdr-entry.cwl"))
    assert echo(message="hello") == {"out": "CONFIGVAR=hello\n"}


@needs_docker
def test_empty_file_creation() -> None:
    """An empty file can be created in InitialWorkingDirectory."""
    err_code, _, _ = get_main_output([get_data("tests/wf/iwdr-empty.cwl")])
    assert err_code == 0


def test_passthrough_successive(tmp_path: Path) -> None:
    """An empty file can be successively passed through a subdir of InitialWorkingDirectory."""
    err_code, _, _ = get_main_output(
        [
            "--outdir",
            str(tmp_path),
            get_data("tests/wf/iwdr-passthrough-successive.cwl"),
        ]
    )
    assert err_code == 0
    children = sorted(tmp_path.glob("*"))  # This input directory should be left pristine.
    assert len(children) == 1
    subdir = tmp_path / children[0]
    assert len(sorted(subdir.glob("*"))) == 1
    assert (subdir / "file").exists()


@needs_docker
def test_directory_literal_with_real_inputs_inside(tmp_path: Path) -> None:
    """Cope with unmoveable files in the output directory created by Docker+IWDR."""
    err_code, _, _ = get_main_output(
        [
            "--out",
            str(tmp_path),
            get_data("tests/iwdr_dir_literal_real_file.cwl"),
            "--example={}".format(get_data("tests/__init__.py")),
        ]
    )
    assert err_code == 0


def test_bad_listing_expression(tmp_path: Path) -> None:
    """Confirm better error message for bad listing expression."""
    err_code, _, stderr = get_main_output(
        [
            "--out",
            str(tmp_path),
            get_data("tests/iwdr_bad_expr.cwl"),
            "--example={}".format(get_data("tests/__init__.py")),
        ]
    )
    stderr = re.sub(r"\s\s+", " ", stderr)
    assert (
        "Expression in a 'InitialWorkdirRequirement.listing' field must return "
        "a list containing zero or more of: File or Directory objects; Dirent "
        "objects. Got 42 among the results" in stderr
    )
    assert err_code == 1


@needs_docker
def test_iwdr_permutations(tmp_path_factory: Any) -> None:
    misc = tmp_path_factory.mktemp("misc")
    fifth = misc / "fifth"
    fifth.mkdir()
    sixth = misc / "sixth"
    sixth.mkdir()
    seventh = misc / "seventh"
    seventh.mkdir()
    eighth = misc / "eighth"
    eighth.mkdir()
    first = misc / "first"
    first.touch()
    second = misc / "second"
    second.touch()
    third = misc / "third"
    third.touch()
    fourth = misc / "fourth"
    fourth.touch()
    eleventh = misc / "eleventh"
    eleventh.touch()
    twelfth = misc / "twelfth"
    twelfth.touch()
    outdir = str(tmp_path_factory.mktemp("outdir"))
    err_code, stdout, _ = get_main_output(
        [
            "--outdir",
            outdir,
            "--debug",
            get_data("tests/wf/iwdr_permutations.cwl"),
            "--first",
            str(first),
            "--second",
            str(second),
            "--third",
            str(third),
            "--fourth",
            str(fourth),
            "--fifth",
            str(fifth),
            "--sixth",
            str(sixth),
            "--seventh",
            str(seventh),
            "--eighth",
            str(eighth),
            "--eleventh",
            str(eleventh),
            "--eleventh",
            str(twelfth),
        ]
    )
    assert err_code == 0
    log = json.loads(stdout)["log"]
    with open(log["path"]) as log_h:
        log_text = log_h.read()
    assert log["checksum"] == "sha1$bc51ebb3f65ca44282789dd1e6de9747d8abe75f", log_text


def test_iwdr_permutations_readonly(tmp_path_factory: Any) -> None:
    """Confirm that readonly input files are properly made writable."""
    misc = tmp_path_factory.mktemp("misc")
    fifth = misc / "fifth"
    fifth.mkdir()
    sixth = misc / "sixth"
    sixth.mkdir()
    fifth_file = fifth / "bar"
    fifth_dir = fifth / "foo"
    fifth_file.touch()
    fifth_dir.mkdir()
    sixth = tmp_path_factory.mktemp("sixth")
    first = misc / "first"
    first.touch()
    second = misc / "second"
    second.touch()
    outdir = str(tmp_path_factory.mktemp("outdir"))
    for entry in [first, second, fifth, sixth, fifth_file, fifth_dir]:
        mode = entry.stat().st_mode
        ro_mask = 0o777 ^ (S_IWRITE | S_IWGRP | S_IWOTH)
        entry.chmod(mode & ro_mask)
    assert (
        main(
            [
                "--no-container",
                "--debug",
                "--leave-outputs",
                "--outdir",
                outdir,
                get_data("tests/wf/iwdr_permutations_nocontainer.cwl"),
                "--first",
                str(first),
                "--second",
                str(second),
                "--fifth",
                str(fifth),
                "--sixth",
                str(sixth),
            ]
        )
        == 0
    )
    for entry in [first, second, fifth, sixth, fifth_file, fifth_dir]:
        try:
            mode = entry.stat().st_mode
            entry.chmod(mode | S_IWRITE)
        except PermissionError:
            pass


@needs_docker
def test_iwdr_permutations_inplace(tmp_path_factory: Any) -> None:
    misc = tmp_path_factory.mktemp("misc")
    fifth = misc / "fifth"
    fifth.mkdir()
    sixth = misc / "sixth"
    sixth.mkdir()
    seventh = misc / "seventh"
    seventh.mkdir()
    eighth = misc / "eighth"
    eighth.mkdir()
    first = misc / "first"
    first.touch()
    second = misc / "second"
    second.touch()
    third = misc / "third"
    third.touch()
    fourth = misc / "fourth"
    fourth.touch()
    eleventh = misc / "eleventh"
    eleventh.touch()
    twelfth = misc / "twelfth"
    twelfth.touch()
    outdir = str(tmp_path_factory.mktemp("outdir"))
    err_code, stdout, _ = get_main_output(
        [
            "--outdir",
            outdir,
            "--enable-ext",
            "--overrides",
            get_data("tests/wf/iwdr_permutations_inplace.yml"),
            get_data("tests/wf/iwdr_permutations.cwl"),
            "--first",
            str(first),
            "--second",
            str(second),
            "--third",
            str(third),
            "--fourth",
            str(fourth),
            "--fifth",
            str(fifth),
            "--sixth",
            str(sixth),
            "--seventh",
            str(seventh),
            "--eighth",
            str(eighth),
            "--eleventh",
            str(eleventh),
            "--eleventh",
            str(twelfth),
        ]
    )
    assert err_code == 0
    log = json.loads(stdout)["log"]
    with open(log["path"]) as log_h:
        log_text = log_h.read()
    assert log["checksum"] == "sha1$bc51ebb3f65ca44282789dd1e6de9747d8abe75f", log_text


@needs_singularity
def test_iwdr_permutations_singularity(
    tmp_path_factory: pytest.TempPathFactory, monkeypatch: pytest.MonkeyPatch
) -> None:
    misc = tmp_path_factory.mktemp("misc")
    fifth = misc / "fifth"
    fifth.mkdir()
    sixth = misc / "sixth"
    sixth.mkdir()
    seventh = misc / "seventh"
    seventh.mkdir()
    eighth = misc / "eighth"
    eighth.mkdir()
    first = misc / "first"
    first.touch()
    second = misc / "second"
    second.touch()
    third = misc / "third"
    third.touch()
    fourth = misc / "fourth"
    fourth.touch()
    eleventh = misc / "eleventh"
    eleventh.touch()
    twelfth = misc / "twelfth"
    twelfth.touch()
    outdir = str(tmp_path_factory.mktemp("outdir"))
    singularity_dir = str(tmp_path_factory.mktemp("singularity"))
    with monkeypatch.context() as m:
        m.setenv("CWL_SINGULARITY_CACHE", singularity_dir)
        err_code, stdout, _ = get_main_output(
            [
                "--outdir",
                outdir,
                "--debug",
                "--singularity",
                get_data("tests/wf/iwdr_permutations.cwl"),
                "--first",
                str(first),
                "--second",
                str(second),
                "--third",
                str(third),
                "--fourth",
                str(fourth),
                "--fifth",
                str(fifth),
                "--sixth",
                str(sixth),
                "--seventh",
                str(seventh),
                "--eighth",
                str(eighth),
                "--eleventh",
                str(eleventh),
                "--eleventh",
                str(twelfth),
            ]
        )
        assert err_code == 0
        log = json.loads(stdout)["log"]
        with open(log["path"]) as log_h:
            log_text = log_h.read()
        assert log["checksum"] == "sha1$bc51ebb3f65ca44282789dd1e6de9747d8abe75f", log_text


@needs_singularity
def test_iwdr_permutations_singularity_inplace(
    tmp_path_factory: pytest.TempPathFactory, monkeypatch: pytest.MonkeyPatch
) -> None:
    """IWDR tests using --singularity and a forced InplaceUpdateRequirement."""
    misc = tmp_path_factory.mktemp("misc")
    fifth = misc / "fifth"
    fifth.mkdir()
    sixth = misc / "sixth"
    sixth.mkdir()
    seventh = misc / "seventh"
    seventh.mkdir()
    eighth = misc / "eighth"
    eighth.mkdir()
    first = misc / "first"
    first.touch()
    second = misc / "second"
    second.touch()
    third = misc / "third"
    third.touch()
    fourth = misc / "fourth"
    fourth.touch()
    eleventh = misc / "eleventh"
    eleventh.touch()
    twelfth = misc / "twelfth"
    twelfth.touch()
    outdir = str(tmp_path_factory.mktemp("outdir"))
    singularity_dir = str(tmp_path_factory.mktemp("singularity"))
    with monkeypatch.context() as m:
        m.setenv("CWL_SINGULARITY_CACHE", singularity_dir)
        assert (
            main(
                [
                    "--outdir",
                    outdir,
                    "--singularity",
                    "--enable-ext",
                    "--enable-dev",
                    "--overrides",
                    get_data("tests/wf/iwdr_permutations_inplace.yml"),
                    get_data("tests/wf/iwdr_permutations.cwl"),
                    "--first",
                    str(first),
                    "--second",
                    str(second),
                    "--third",
                    str(third),
                    "--fourth",
                    str(fourth),
                    "--fifth",
                    str(fifth),
                    "--sixth",
                    str(sixth),
                    "--seventh",
                    str(seventh),
                    "--eighth",
                    str(eighth),
                    "--eleventh",
                    str(eleventh),
                    "--eleventh",
                    str(twelfth),
                ]
            )
            == 0
        )


def test_iwdr_writable_secondaryfiles(tmp_path: Path) -> None:
    """Confirm that a writable primary file and its secondary files are still co-located."""
    err_code, _, _ = get_main_output(
        [
            "--outdir",
            str(tmp_path),
            get_data("tests/iwdr_writable_secfile.cwl"),
            get_data("tests/iwdr_writable_secfile-inputs.json"),
        ]
    )
    assert err_code == 0



from io import StringIO
from pathlib import Path

from cwltool.main import main

from .util import get_data


def test_empty_input(tmp_path: Path) -> None:
    """Affirm that an empty input works."""
    empty_json = "{}"
    empty_input = StringIO(empty_json)

    params = [
        "--outdir",
        str(tmp_path),
        get_data("tests/wf/no-parameters-echo.cwl"),
        "-",
    ]

    try:
        assert main(params, stdin=empty_input) == 0
    except SystemExit as err:
        assert err.code == 0



import prov.identifier
from _typeshed import Incomplete
from prov.constants import *
from prov.model import DEFAULT_NAMESPACES as DEFAULT_NAMESPACES
from prov.model import sorted_attributes as sorted_attributes

logger: Incomplete
FULL_NAMES_MAP: Incomplete
FULL_PROV_RECORD_IDS_MAP: Incomplete
XML_XSD_URI: str

class ProvXMLException(prov.Error): ...

class ProvXMLSerializer(prov.serializers.Serializer):
    def serialize(self, stream, force_types: bool = ..., **kwargs) -> None: ...
    def serialize_bundle(
        self, bundle, element: Incomplete | None = ..., force_types: bool = ...
    ): ...
    def deserialize(self, stream, **kwargs): ...
    def deserialize_subtree(self, xml_doc, bundle): ...

def xml_qname_to_QualifiedName(element, qname_str): ...



"""Visualize a CWL workflow."""

from collections.abc import Iterator
from importlib.resources import files
from typing import cast
from urllib.parse import urlparse

import pydot
import rdflib
from packaging.version import Version

if Version(pydot.__version__) > Version("3.0"):
    quote_id_if_necessary = pydot.quote_id_if_necessary
else:
    quote_id_if_necessary = pydot.quote_if_necessary  # type: ignore[attr-defined]


def _get_inner_edges_query() -> str:
    return files("cwltool").joinpath("rdfqueries/get_inner_edges.sparql").read_text()


def _get_input_edges_query() -> str:
    return files("cwltool").joinpath("rdfqueries/get_input_edges.sparql").read_text()


def _get_output_edges_query() -> str:
    return files("cwltool").joinpath("rdfqueries/get_output_edges.sparql").read_text()


def _get_root_query() -> str:
    return files("cwltool").joinpath("rdfqueries/get_root.sparql").read_text()


class CWLViewer:
    """Produce similar images with the https://github.com/common-workflow-language/cwlviewer."""

    def __init__(self, rdf_description: str):
        """Create a viewer object based on the rdf description of the workflow."""
        self._dot_graph: pydot.Graph = CWLViewer._init_dot_graph()
        self._rdf_graph: rdflib.graph.Graph = self._load_cwl_graph(rdf_description)
        self._root_graph_uri: str = self._get_root_graph_uri()
        self._set_inner_edges()
        self._set_input_edges()
        self._set_output_edges()

    def _load_cwl_graph(self, rdf_description: str) -> rdflib.graph.Graph:
        rdf_graph = rdflib.Graph()
        rdf_graph.parse(data=rdf_description, format="n3")
        return rdf_graph

    def _set_inner_edges(self) -> None:
        get_inner_edges_query = _get_inner_edges_query()
        inner_edges = cast(
            Iterator[rdflib.query.ResultRow],
            self._rdf_graph.query(
                get_inner_edges_query, initBindings={"root_graph": self._root_graph_uri}
            ),
        )  # ResultRow because the query is of type SELECT
        for inner_edge_row in inner_edges:
            source_label = (
                inner_edge_row["source_label"]
                if inner_edge_row["source_label"] is not None
                else urlparse(inner_edge_row["source_step"]).fragment
            )
            # Node color and style depend on class
            source_color = (
                "#F3CEA1"
                if inner_edge_row["source_step_class"].endswith("Workflow")
                else "lightgoldenrodyellow"
            )
            source_style = (
                "dashed" if inner_edge_row["source_step_class"].endswith("Operation") else "filled"
            )
            n = pydot.Node(
                "",
                fillcolor=source_color,
                style=source_style,
                label=source_label,
                shape="record",
            )
            n.set_name(str(inner_edge_row["source_step"]))
            self._dot_graph.add_node(n)
            target_label = (
                inner_edge_row["target_label"]
                if inner_edge_row["target_label"] is not None
                else urlparse(inner_edge_row["target_step"]).fragment
            )

            target_color = (
                "#F3CEA1"
                if inner_edge_row["target_step_class"].endswith("Workflow")
                else "lightgoldenrodyellow"
            )
            target_style = (
                "dashed" if inner_edge_row["target_step_class"].endswith("Operation") else "filled"
            )
            n = pydot.Node(
                "",
                fillcolor=target_color,
                style=target_style,
                label=target_label,
                shape="record",
            )
            n.set_name(str(inner_edge_row["target_step"]))
            self._dot_graph.add_node(n)
            self._dot_graph.add_edge(
                pydot.Edge(
                    quote_id_if_necessary(str(inner_edge_row["source_step"])),
                    quote_id_if_necessary(str(inner_edge_row["target_step"])),
                )
            )

    def _set_input_edges(self) -> None:
        get_input_edges_query = _get_input_edges_query()
        inputs_subgraph = pydot.Subgraph(graph_name="cluster_inputs")
        self._dot_graph.add_subgraph(inputs_subgraph)
        inputs_subgraph.set("rank", "same")
        inputs_subgraph.set("style", "dashed")
        inputs_subgraph.set("label", "Workflow Inputs")

        input_edges = cast(
            Iterator[rdflib.query.ResultRow],
            self._rdf_graph.query(
                get_input_edges_query, initBindings={"root_graph": self._root_graph_uri}
            ),
        )  # ResultRow because the query is of type SELECT
        for input_row in input_edges:
            n = pydot.Node(
                "",
                fillcolor="#94DDF4",
                style="filled",
                label=urlparse(input_row["input"]).fragment,
                shape="record",
            )
            n.set_name(str(input_row["input"]))
            inputs_subgraph.add_node(n)
            self._dot_graph.add_edge(
                pydot.Edge(
                    quote_id_if_necessary(str(input_row["input"])),
                    quote_id_if_necessary(str(input_row["step"])),
                )
            )

    def _set_output_edges(self) -> None:
        get_output_edges = _get_output_edges_query()
        outputs_graph = pydot.Subgraph(graph_name="cluster_outputs")
        self._dot_graph.add_subgraph(outputs_graph)
        outputs_graph.set("rank", "same")
        outputs_graph.set("style", "dashed")
        outputs_graph.set("label", "Workflow Outputs")
        outputs_graph.set("labelloc", "b")
        output_edges = cast(
            Iterator[rdflib.query.ResultRow],
            self._rdf_graph.query(
                get_output_edges, initBindings={"root_graph": self._root_graph_uri}
            ),
        )  # ResultRow because the query is of type SELECT
        for output_edge_row in output_edges:
            n = pydot.Node(
                "",
                fillcolor="#94DDF4",
                style="filled",
                label=urlparse(output_edge_row["output"]).fragment,
                shape="record",
            )
            n.set_name(str(output_edge_row["output"]))
            outputs_graph.add_node(n)
            self._dot_graph.add_edge(
                pydot.Edge(
                    quote_id_if_necessary(output_edge_row["step"]),
                    quote_id_if_necessary(output_edge_row["output"]),
                )
            )

    def _get_root_graph_uri(self) -> rdflib.term.Identifier:
        get_root_query = _get_root_query()
        root = cast(
            list[rdflib.query.ResultRow],
            list(
                self._rdf_graph.query(
                    get_root_query,
                )
            ),
        )  # ResultRow because the query is of type SELECT
        if len(root) != 1:
            raise RuntimeError(
                "Cannot identify root workflow! Notice that only Workflows can be visualized"
            )

        workflow = root[0]["workflow"]
        return workflow

    @classmethod
    def _init_dot_graph(cls) -> pydot.Graph:
        graph = pydot.Graph(graph_type="digraph", simplify=False)
        graph.set("bgcolor", "#eeeeee")
        graph.set("clusterrank", "local")
        graph.set("labelloc", "bottom")
        graph.set("labelloc", "bottom")
        graph.set("labeljust", "right")

        return graph

    def get_dot_graph(self) -> pydot.Graph:
        """Get the dot graph object."""
        return self._dot_graph

    def dot(self) -> str:
        """Get the graph as graphviz."""
        return str(self._dot_graph.to_string())



from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class DOAP(DefinedNamespace):
    audience: URIRef
    blog: URIRef
    browse: URIRef
    category: URIRef
    created: URIRef
    description: URIRef
    developer: URIRef
    documenter: URIRef
    helper: URIRef
    implements: URIRef
    language: URIRef
    license: URIRef
    location: URIRef
    maintainer: URIRef
    module: URIRef
    name: URIRef
    os: URIRef
    platform: URIRef
    release: URIRef
    repository: URIRef
    repositoryOf: URIRef
    revision: URIRef
    screenshots: URIRef
    shortdesc: URIRef
    tester: URIRef
    translator: URIRef
    vendor: URIRef
    wiki: URIRef
    ArchRepository: URIRef
    BKRepository: URIRef
    BazaarBranch: URIRef
    CVSRepository: URIRef
    DarcsRepository: URIRef
    GitBranch: URIRef
    GitRepository: URIRef
    HgRepository: URIRef
    Project: URIRef
    Repository: URIRef
    SVNRepository: URIRef
    Specification: URIRef
    Version: URIRef
    homepage: URIRef



import subprocess
import sys

import pytest

from cwltool.main import main

from .util import get_data, get_main_output


def test_rdf_print() -> None:
    assert main(["--print-rdf", get_data("tests/wf/hello_single_tool.cwl")]) == 0


def test_rdf_print_inputs_and_defaults() -> None:
    error_code, stdout, stderr = get_main_output(
        ["--print-rdf", get_data("tests/wf/revsort.cwl"), get_data("tests/wf/revsort-job.json")]
    )
    assert error_code == 0, stderr
    assert "revsort.cwl#workflow_input> rdf:value " in stdout
    assert (
        'tests/wf/revsort.cwl#workflow_input> rdfs:comment "The input file to be processed."'
        not in stdout
    )


def test_rdf_print_unicode(monkeypatch: pytest.MonkeyPatch) -> None:
    """Force ASCII encoding but load UTF file with --print-rdf."""
    with monkeypatch.context() as m:
        m.setenv("LC_ALL", "C")
        params = [
            sys.executable,
            "-m",
            "cwltool",
            "--print-rdf",
            get_data("tests/utf_doc_example.cwl"),
        ]
        assert subprocess.check_call(params) == 0



from ..recipe import Recipe
from .base import ParserBase as ParserBase

class DockerParser(ParserBase):
    name: str
    def __init__(self, filename: str = ..., load: bool = ...) -> None: ...
    def parse(self) -> dict[str, Recipe]: ...



"""Tests --validation."""

import io
import logging
import re

from .util import get_data, get_main_output


def test_validate_graph_with_no_default() -> None:
    """Ensure that --validate works on $graph docs that lack a main/#main."""
    exit_code, stdout, stderr = get_main_output(
        ["--validate", get_data("tests/wf/packed_no_main.cwl")]
    )
    assert exit_code == 0
    assert "packed_no_main.cwl#echo is valid CWL" in stdout
    assert "packed_no_main.cwl#cat is valid CWL" in stdout
    assert "packed_no_main.cwl#collision is valid CWL" in stdout
    assert "tests/wf/packed_no_main.cwl is valid CWL" in stdout


def test_validate_with_valid_input_object() -> None:
    """Ensure that --validate with a valid input object."""
    exit_code, stdout, stderr = get_main_output(
        [
            "--validate",
            get_data("tests/wf/1st-workflow.cwl"),
            "--inp",
            get_data("tests/wf/1st-workflow.cwl"),
            "--ex",
            "FOO",
        ]
    )
    assert exit_code == 0
    assert "tests/wf/1st-workflow.cwl is valid CWL. No errors detected in the inputs." in stdout


def test_validate_with_invalid_input_object() -> None:
    """Ensure that --validate with an invalid input object."""
    exit_code, stdout, stderr = get_main_output(
        [
            "--validate",
            get_data("tests/wf/1st-workflow.cwl"),
            get_data("tests/wf/1st-workflow_bad_inputs.yml"),
        ]
    )
    assert exit_code == 1
    stdout = re.sub(r"\s\s+", " ", stdout)
    assert "Invalid job input record" in stdout
    assert (
        "tests/wf/1st-workflow_bad_inputs.yml:2:1: * the 'ex' field is not "
        "valid because the value is not string" in stdout
    )
    assert (
        "tests/wf/1st-workflow_bad_inputs.yml:1:1: * the 'inp' field is not "
        "valid because is not a dict. Expected a File object." in stdout
    )


def test_validate_quiet() -> None:
    """Ensure that --validate --quiet prints the correct amount of information."""
    exit_code, stdout, stderr = get_main_output(
        [
            "--validate",
            "--quiet",
            get_data("tests/CometAdapter.cwl"),
        ]
    )
    assert exit_code == 0
    stdout = re.sub(r"\s\s+", " ", stdout)
    assert "INFO" not in stdout
    assert "INFO" not in stderr
    assert "tests/CometAdapter.cwl:10:3: object id" in stdout
    assert "tests/CometAdapter.cwl#out' previously defined" in stdout


def test_validate_no_warnings() -> None:
    """Ensure that --validate --no-warnings doesn't print any warnings."""
    exit_code, stdout, stderr = get_main_output(
        [
            "--validate",
            "--no-warnings",
            get_data("tests/CometAdapter.cwl"),
        ]
    )
    assert exit_code == 0
    stdout = re.sub(r"\s\s+", " ", stdout)
    stderr = re.sub(r"\s\s+", " ", stderr)
    assert "INFO" not in stdout
    assert "INFO" not in stderr
    assert "WARNING" not in stdout
    assert "WARNING" not in stderr
    assert "tests/CometAdapter.cwl:9:3: object id" not in stdout
    assert "tests/CometAdapter.cwl:9:3: object id" not in stderr
    assert "tests/CometAdapter.cwl#out' previously defined" not in stdout
    assert "tests/CometAdapter.cwl#out' previously defined" not in stderr


def test_validate_custom_logger() -> None:
    """Custom log handling test."""
    custom_log = io.StringIO()
    handler = logging.StreamHandler(custom_log)
    handler.setLevel(logging.DEBUG)
    exit_code, stdout, stderr = get_main_output(
        [
            "--validate",
            get_data("tests/CometAdapter.cwl"),
        ],
        logger_handler=handler,
    )
    custom_log_text = custom_log.getvalue()
    assert exit_code == 0
    custom_log_text = re.sub(r"\s\s+", " ", custom_log_text)
    stdout = re.sub(r"\s\s+", " ", stdout)
    stderr = re.sub(r"\s\s+", " ", stderr)
    assert "INFO" not in stdout
    assert "INFO" not in stderr
    assert "INFO" in custom_log_text
    assert "WARNING" not in stdout
    assert "WARNING" not in stderr
    assert "WARNING" in custom_log_text
    assert "tests/CometAdapter.cwl:10:3: object id" not in stdout
    assert "tests/CometAdapter.cwl:10:3: object id" not in stderr
    assert "tests/CometAdapter.cwl:10:3: object id" in custom_log_text
    assert "tests/CometAdapter.cwl#out' previously defined" not in stdout
    assert "tests/CometAdapter.cwl#out' previously defined" not in stderr
    assert "tests/CometAdapter.cwl#out' previously defined" in custom_log_text



from typing import Any, Tuple, Union
from urllib.parse import ParseResult
from uuid import NAMESPACE_URL, UUID

SCHEME: str

def is_arcp_uri(uri: str) -> bool: ...
def parse_arcp(uri: str) -> "ARCPParseResult": ...
def urlparse(uri: str) -> Union["ARCPParseResult", ParseResult]: ...

class ARCPParseResult(ParseResult):
    def __init__(self, *args: Any) -> None: ...
    @property
    def prefix(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def uuid(self) -> UUID: ...
    @property
    def ni(self) -> str | None: ...
    def ni_uri(self, authority: str = ...) -> str | None: ...
    def nih_uri(self) -> str | None: ...
    def ni_well_known(self, base: str = ...) -> str | None: ...
    @property
    def hash(self) -> tuple[str, str] | None: ...



"""Stores class definition of ResearchObject and WritableBagFile."""

import datetime
import hashlib
import os
import shutil
import tempfile
import urllib
import uuid
from collections.abc import MutableMapping, MutableSequence
from pathlib import Path, PurePosixPath
from socket import getfqdn
from typing import IO, TYPE_CHECKING, Any, Optional, cast

import prov.model as provM
from prov.model import ProvDocument

from ..loghandler import _logger
from ..stdfsaccess import StdFsAccess
from ..utils import (
    CWLObjectType,
    CWLOutputType,
    create_tmp_dir,
    local_path,
    posix_path,
    versionstring,
)
from . import Aggregate, Annotation, AuthoredBy, _valid_orcid, _whoami, checksum_copy
from .provenance_constants import (
    ACCOUNT_UUID,
    CWLPROV,
    CWLPROV_VERSION,
    DATA,
    ENCODING,
    FOAF,
    LOGS,
    METADATA,
    ORCID,
    PROVENANCE,
    SCHEMA,
    SHA1,
    SHA256,
    SHA512,
    SNAPSHOT,
    TEXT_PLAIN,
    USER_UUID,
    UUID,
    WORKFLOW,
    Hasher,
)

if TYPE_CHECKING:
    from .provenance_profile import ProvenanceProfile  # pylint: disable=unused-import


class ResearchObject:
    """CWLProv Research Object."""

    def __init__(
        self,
        fsaccess: StdFsAccess,
        temp_prefix_ro: str = "tmp",
        orcid: str = "",
        full_name: str = "",
    ) -> None:
        """Initialize the ResearchObject."""
        self.temp_prefix = temp_prefix_ro
        self.orcid = "" if not orcid else _valid_orcid(orcid)
        self.full_name = full_name
        self.folder = create_tmp_dir(temp_prefix_ro)
        self.closed = False
        # map of filename "data/de/alsdklkas": 12398123 bytes
        self.bagged_size: dict[str, int] = {}
        self.tagfiles: set[str] = set()
        self._file_provenance: dict[str, Aggregate] = {}
        self._external_aggregates: list[Aggregate] = []
        self.annotations: list[Annotation] = []
        self._content_types: dict[str, str] = {}
        self.fsaccess = fsaccess
        # These should be replaced by generate_prov_doc when workflow/run IDs are known:
        self.engine_uuid = f"urn:uuid:{uuid.uuid4()}"
        self.ro_uuid = uuid.uuid4()
        self.base_uri = f"arcp://uuid,{self.ro_uuid}/"
        self.cwltool_version = f"cwltool {versionstring().split()[-1]}"
        self.has_manifest = False
        self.relativised_input_object: CWLObjectType = {}

        self._initialize()
        _logger.debug("[provenance] Temporary research object: %s", self.folder)

    def initialize_provenance(
        self,
        full_name: str,
        host_provenance: bool,
        user_provenance: bool,
        orcid: str,
        fsaccess: StdFsAccess,
        run_uuid: uuid.UUID | None = None,
    ) -> "ProvenanceProfile":
        """
        Provide a provenance profile initialization hook function.

        Allows overriding the default strategy to define the
        provenance profile concepts and associations to extend
        details as needed.
        """
        from .provenance_profile import ProvenanceProfile

        return ProvenanceProfile(
            research_object=self,
            full_name=full_name,
            host_provenance=host_provenance,
            user_provenance=user_provenance,
            orcid=orcid,
            fsaccess=fsaccess,
            run_uuid=run_uuid,
        )

    def self_check(self) -> None:
        """Raise ValueError if this RO is closed."""
        if self.closed:
            raise ValueError(
                "This ResearchObject has already been closed and is not "
                "available for further manipulation."
            )

    def __str__(self) -> str:
        """Represent this RO as a string."""
        return f"ResearchObject <{self.ro_uuid}> in <{self.folder}>"

    def _initialize(self) -> None:
        """Initialize the bagit folder structure."""
        for research_obj_folder in (
            METADATA,
            DATA,
            WORKFLOW,
            SNAPSHOT,
            PROVENANCE,
            LOGS,
        ):
            os.makedirs(os.path.join(self.folder, research_obj_folder))
        self._initialize_bagit()

    def _initialize_bagit(self) -> None:
        """Write fixed bagit header."""
        self.self_check()
        bagit = os.path.join(self.folder, "bagit.txt")
        # encoding: always UTF-8 (although ASCII would suffice here)
        with open(bagit, "w", encoding=ENCODING, newline="\n") as bag_it_file:
            # TODO: \n or \r\n ?
            bag_it_file.write("BagIt-Version: 0.97\n")
            bag_it_file.write(f"Tag-File-Character-Encoding: {ENCODING}\n")

    def resolve_user(self) -> tuple[str, str]:
        """
        Provide a user provenance hook function.

        Allows overriding the default strategy to retrieve user provenance
        in case the calling code can provide a better resolution.
        The function must return a tuple of the (username, fullname)
        that identifies the user. This user will be applied on top
        to any provided ORCID or fullname by agent association.
        """
        return _whoami()

    def user_provenance(self, document: ProvDocument) -> None:
        """Add the user provenance."""
        self.self_check()
        username, fullname = self.resolve_user()

        if not self.full_name:
            self.full_name = fullname

        document.add_namespace(UUID)
        document.add_namespace(ORCID)
        document.add_namespace(FOAF)
        account = document.agent(
            ACCOUNT_UUID,
            {
                provM.PROV_TYPE: FOAF["OnlineAccount"],
                provM.PROV_LABEL: username,
                FOAF["accountName"]: username,
            },
        )

        user = document.agent(
            self.orcid or USER_UUID,
            [
                (provM.PROV_TYPE, SCHEMA["Person"]),
                (provM.PROV_TYPE, provM.PROV["Person"]),
                (provM.PROV_LABEL, self.full_name),
                (FOAF["name"], self.full_name),
                (FOAF["account"], account),
                (SCHEMA["name"], self.full_name),
            ],
        )
        # cwltool may be started on the shell (directly by user),
        # by shell script (indirectly by user)
        # or from a different program
        #   (which again is launched by any of the above)
        #
        # We can't tell in which way, but ultimately we're still
        # acting in behalf of that user (even if we might
        # get their name wrong!)
        document.actedOnBehalfOf(account, user)

    def resolve_host(self) -> tuple[str, str]:
        """
        Provide a host provenance hook function.

        Allows overriding the default strategy to retrieve host provenance
        in case the calling code can provide a better resolution.
        The function must return a tuple of the (fqdn, uri) that identifies the host.
        """
        fqdn = getfqdn()
        return fqdn, fqdn  # allow for (fqdn, uri) to be distinct, but the same by default

    def host_provenance(self, document: ProvDocument) -> None:
        """Record host provenance."""
        document.add_namespace(CWLPROV)
        document.add_namespace(UUID)
        document.add_namespace(FOAF)

        hostname, uri = self.resolve_host()
        # won't have a foaf:accountServiceHomepage for unix hosts, but
        # we can at least provide hostname
        document.agent(
            ACCOUNT_UUID,
            {
                provM.PROV_TYPE: FOAF["OnlineAccount"],
                provM.PROV_LOCATION: uri,
                CWLPROV["hostname"]: hostname,
            },
        )

    def add_tagfile(self, path: str, timestamp: datetime.datetime | None = None) -> None:
        """Add tag files to our research object."""
        self.self_check()
        checksums = {}
        # Read file to calculate its checksum
        if os.path.isdir(path):
            return
            # FIXME: do the right thing for directories
        with open(path, "rb") as tag_file:
            # FIXME: Should have more efficient open_tagfile() that
            # does all checksums in one go while writing through,
            # adding checksums after closing.
            # Below probably OK for now as metadata files
            # are not too large..?

            checksums[SHA1] = checksum_copy(tag_file, hasher=hashlib.sha1)

            tag_file.seek(0)
            checksums[SHA256] = checksum_copy(tag_file, hasher=hashlib.sha256)

            tag_file.seek(0)
            checksums[SHA512] = checksum_copy(tag_file, hasher=hashlib.sha512)

        rel_path = posix_path(os.path.relpath(path, self.folder))
        self.tagfiles.add(rel_path)
        self.add_to_manifest(rel_path, checksums)
        if timestamp is not None:
            self._file_provenance[rel_path] = {
                "createdOn": timestamp.isoformat(),
                "uri": None,
                "bundledAs": None,
                "mediatype": None,
                "conformsTo": None,
            }

    def _ro_aggregates(self) -> list[Aggregate]:
        """Gather dictionary of files to be added to the manifest."""

        def guess_mediatype(
            rel_path: str,
        ) -> tuple[str | None, str | list[str] | None]:
            """Return the mediatypes."""
            media_types: dict[str | None, str] = {
                # Adapted from
                # https://w3id.org/bundle/2014-11-05/#media-types
                "txt": TEXT_PLAIN,
                "ttl": 'text/turtle; charset="UTF-8"',
                "rdf": "application/rdf+xml",
                "json": "application/json",
                "jsonld": "application/ld+json",
                "xml": "application/xml",
                ##
                "cwl": 'text/x+yaml; charset="UTF-8"',
                "provn": 'text/provenance-notation; charset="UTF-8"',
                "nt": "application/n-triples",
            }
            conforms_to: dict[str | None, str] = {
                "provn": "http://www.w3.org/TR/2013/REC-prov-n-20130430/",
                "cwl": "https://w3id.org/cwl/",
            }

            prov_conforms_to: dict[str, str] = {
                "provn": "http://www.w3.org/TR/2013/REC-prov-n-20130430/",
                "rdf": "http://www.w3.org/TR/2013/REC-prov-o-20130430/",
                "ttl": "http://www.w3.org/TR/2013/REC-prov-o-20130430/",
                "nt": "http://www.w3.org/TR/2013/REC-prov-o-20130430/",
                "jsonld": "http://www.w3.org/TR/2013/REC-prov-o-20130430/",
                "xml": "http://www.w3.org/TR/2013/NOTE-prov-xml-20130430/",
                "json": "http://www.w3.org/Submission/2013/SUBM-prov-json-20130424/",
            }

            extension: str | None = rel_path.rsplit(".", 1)[-1].lower()
            if extension == rel_path:
                # No ".", no extension
                extension = None

            mediatype: str | None = media_types.get(extension, None)
            conformsTo: str | list[str] | None = conforms_to.get(extension, None)
            # TODO: Open CWL file to read its declared "cwlVersion", e.g.
            # cwlVersion = "v1.0"

            if rel_path.startswith(posix_path(PROVENANCE)) and extension in prov_conforms_to:
                if ".cwlprov" in rel_path:
                    # Our own!
                    conformsTo = [
                        prov_conforms_to[extension],
                        CWLPROV_VERSION,
                    ]
                else:
                    # Some other PROV
                    # TODO: Recognize ProvOne etc.
                    conformsTo = prov_conforms_to[extension]
            return (mediatype, conformsTo)

        aggregates: list[Aggregate] = []
        for path in self.bagged_size.keys():
            temp_path = PurePosixPath(path)
            folder = temp_path.parent
            filename = temp_path.name

            # NOTE: Here we end up aggregating the abstract
            # data items by their sha1 hash, so that it matches
            # the entity() in the prov files.

            # TODO: Change to nih:sha-256; hashes
            #  https://tools.ietf.org/html/rfc6920#section-7
            aggregate_dict: Aggregate = {
                "uri": "urn:hash::sha1:" + filename,
                "bundledAs": {
                    # The arcp URI is suitable ORE proxy; local to this Research Object.
                    # (as long as we don't also aggregate it by relative path!)
                    "uri": self.base_uri + path,
                    # relate it to the data/ path
                    "folder": f"/{folder}/",
                    "filename": filename,
                },
            }
            if path in self._file_provenance:
                # Made by workflow run, merge captured provenance
                bundledAs = aggregate_dict["bundledAs"]
                if bundledAs:
                    bundledAs.update(self._file_provenance[path])
                else:
                    aggregate_dict["bundledAs"] = cast(
                        Optional[dict[str, Any]], self._file_provenance[path]
                    )
            else:
                # Probably made outside wf run, part of job object?
                pass
            if path in self._content_types:
                aggregate_dict["mediatype"] = self._content_types[path]

            aggregates.append(aggregate_dict)

        for path in self.tagfiles:
            if not (
                path.startswith(METADATA) or path.startswith(WORKFLOW) or path.startswith(SNAPSHOT)
            ):
                # probably a bagit file
                continue
            if path == str(PurePosixPath(METADATA) / "manifest.json"):
                # Should not really be there yet! But anyway, we won't
                # aggregate it.
                continue

            # These are local paths like metadata/provenance - but
            # we need to relativize them for our current directory for
            # as we are saved in metadata/manifest.json
            mediatype, conformsTo = guess_mediatype(path)
            rel_aggregates: Aggregate = {
                "uri": str(Path(os.pardir) / path),
                "mediatype": mediatype,
                "conformsTo": conformsTo,
            }

            if path in self._file_provenance:
                # Propagate file provenance (e.g. timestamp)
                rel_aggregates.update(self._file_provenance[path])
            elif not path.startswith(SNAPSHOT):
                # make new timestamp?
                (
                    rel_aggregates["createdOn"],
                    rel_aggregates["createdBy"],
                ) = self._self_made()
            aggregates.append(rel_aggregates)
        aggregates.extend(self._external_aggregates)
        return aggregates

    def add_uri(self, uri: str, timestamp: datetime.datetime | None = None) -> Aggregate:
        """Add the given URI to this RO."""
        self.self_check()
        aggr: Aggregate = {"uri": uri}
        aggr["createdOn"], aggr["createdBy"] = self._self_made(timestamp=timestamp)
        self._external_aggregates.append(aggr)
        return aggr

    def add_annotation(
        self, about: str, content: list[str], motivated_by: str = "oa:describing"
    ) -> str:
        """Cheap URI relativize for current directory and /."""
        self.self_check()
        curr = self.base_uri + METADATA + "/"
        content = [c.replace(curr, "").replace(self.base_uri, "../") for c in content]
        uri = uuid.uuid4().urn
        ann: Annotation = {
            "uri": uri,
            "about": about,
            "content": content,
            "oa:motivatedBy": {"@id": motivated_by},
        }
        self.annotations.append(ann)
        return uri

    def _ro_annotations(self) -> list[Annotation]:
        """Append base RO and provenance annotations to the list of annotations."""
        annotations: list[Annotation] = []
        annotations.append(
            {
                "uri": uuid.uuid4().urn,
                "about": self.ro_uuid.urn,
                "content": "/",
                # https://www.w3.org/TR/annotation-vocab/#named-individuals
                "oa:motivatedBy": {"@id": "oa:describing"},
            }
        )

        # How was it run?
        # FIXME: Only primary*
        prov_files = [
            str(PurePosixPath(p).relative_to(METADATA))
            for p in self.tagfiles
            if p.startswith(posix_path(PROVENANCE)) and "/primary." in p
        ]
        annotations.append(
            {
                "uri": uuid.uuid4().urn,
                "about": self.ro_uuid.urn,
                "content": prov_files,
                # Modulation of https://www.w3.org/TR/prov-aq/
                "oa:motivatedBy": {"@id": "http://www.w3.org/ns/prov#has_provenance"},
            }
        )

        # Where is the main workflow?
        annotations.append(
            {
                "uri": uuid.uuid4().urn,
                "about": str(PurePosixPath("..") / WORKFLOW / "packed.cwl"),
                "content": None,
                "oa:motivatedBy": {"@id": "oa:highlighting"},
            }
        )

        annotations.append(
            {
                "uri": uuid.uuid4().urn,
                "about": self.ro_uuid.urn,
                "content": [
                    str(PurePosixPath("..") / WORKFLOW / "packed.cwl"),
                    str(PurePosixPath("..") / WORKFLOW / "primary-job.json"),
                ],
                "oa:motivatedBy": {"@id": "oa:linking"},
            }
        )
        # Add user-added annotations at end
        annotations.extend(self.annotations)
        return annotations

    def _authored_by(self) -> AuthoredBy | None:
        authored_by: AuthoredBy = {}
        """Returns the authoredBy metadata if it was supplied on CLI"""
        if self.orcid:
            authored_by["orcid"] = self.orcid
        if self.full_name:
            authored_by["name"] = self.full_name
            if not self.orcid:
                authored_by["uri"] = USER_UUID

        if authored_by:
            return authored_by
        return None

    def generate_snapshot(self, prov_dep: CWLObjectType) -> None:
        """Copy all of the CWL files to the snapshot/ directory."""
        self.self_check()
        for key, value in prov_dep.items():
            if key == "location" and cast(str, value).split("/")[-1]:
                location = urllib.parse.unquote(cast(str, value))
                filename = location.split("/")[-1]
                path = os.path.join(self.folder, SNAPSHOT, filename)
                filepath = ""
                if "file://" in location:
                    filepath = location[7:]
                else:
                    filepath = location

                # FIXME: What if destination path already exists?
                if os.path.exists(filepath):
                    try:
                        if os.path.isdir(filepath):
                            shutil.copytree(filepath, path)
                        else:
                            shutil.copy(filepath, path)
                        timestamp = datetime.datetime.fromtimestamp(os.path.getmtime(filepath))
                        self.add_tagfile(path, timestamp)
                    except PermissionError:
                        pass  # FIXME: avoids duplicate snapshotting; need better solution
            elif key in ("secondaryFiles", "listing"):
                for files in cast(MutableSequence[CWLObjectType], value):
                    if isinstance(files, MutableMapping):
                        self.generate_snapshot(files)
            else:
                pass

    def has_data_file(self, sha1hash: str) -> bool:
        """Confirm the presence of the given file in the RO."""
        folder = os.path.join(self.folder, DATA, sha1hash[0:2])
        hash_path = os.path.join(folder, sha1hash)
        return os.path.isfile(hash_path)

    def add_data_file(
        self,
        from_fp: IO[Any],
        timestamp: datetime.datetime | None = None,
        content_type: str | None = None,
    ) -> str:
        """Copy inputs to data/ folder."""
        self.self_check()
        tmp_dir, tmp_prefix = os.path.split(self.temp_prefix)
        with tempfile.NamedTemporaryFile(prefix=tmp_prefix, dir=tmp_dir, delete=False) as tmp:
            checksum = checksum_copy(from_fp, tmp)

        # Calculate hash-based file path
        folder = os.path.join(self.folder, DATA, checksum[0:2])
        path = os.path.join(folder, checksum)
        # os.rename assumed safe, as our temp file should
        # be in same file system as our temp folder
        if not os.path.isdir(folder):
            os.makedirs(folder)
        os.rename(tmp.name, path)

        # Relative posix path
        rel_path = posix_path(os.path.relpath(path, self.folder))

        # Register in bagit checksum
        if Hasher == hashlib.sha1:
            self._add_to_bagit(rel_path, sha1=checksum)
        else:
            _logger.warning("[provenance] Unknown hash method %s for bagit manifest", Hasher)
            # Inefficient, bagit support need to checksum again
            self._add_to_bagit(rel_path)
        _logger.debug("[provenance] Added data file %s", path)
        if timestamp is not None:
            createdOn, createdBy = self._self_made(timestamp)
            self._file_provenance[rel_path] = cast(
                Aggregate, {"createdOn": createdOn, "createdBy": createdBy}
            )
        _logger.debug("[provenance] Relative path for data file %s", rel_path)

        if content_type is not None:
            self._content_types[rel_path] = content_type
        return rel_path

    def _self_made(
        self, timestamp: datetime.datetime | None = None
    ) -> tuple[str, dict[str, str]]:  # createdOn, createdBy
        if timestamp is None:
            timestamp = datetime.datetime.now()
        return (
            timestamp.isoformat(),
            {"uri": self.engine_uuid, "name": self.cwltool_version},
        )

    def add_to_manifest(self, rel_path: str, checksums: dict[str, str]) -> None:
        """Add files to the research object manifest."""
        self.self_check()
        if PurePosixPath(rel_path).is_absolute():
            raise ValueError(f"rel_path must be relative: {rel_path}")

        if os.path.commonprefix(["data/", rel_path]) == "data/":
            # payload file, go to manifest
            manifest = "manifest"
            self.has_manifest = True
        else:
            # metadata file, go to tag manifest
            manifest = "tagmanifest"

        # Add checksums to corresponding manifest files
        for method, hash_value in checksums.items():
            # File not in manifest because we bailed out on
            # existence in bagged_size above
            manifestpath = os.path.join(self.folder, f"{manifest}-{method.lower()}.txt")
            # encoding: match Tag-File-Character-Encoding: UTF-8
            with open(manifestpath, "a", encoding=ENCODING, newline="\n") as checksum_file:
                line = f"{hash_value}  {rel_path}\n"
                _logger.debug("[provenance] Added to %s: %s", manifestpath, line)
                checksum_file.write(line)

    def _add_to_bagit(self, rel_path: str, **checksums: str) -> None:
        """Compute file size and checksums and adds to bagit manifest."""
        if PurePosixPath(rel_path).is_absolute():
            raise ValueError(f"rel_path must be relative: {rel_path}")
        lpath = os.path.join(self.folder, local_path(rel_path))
        if not os.path.exists(lpath):
            raise OSError(f"File {rel_path} does not exist within RO: {lpath}")

        if rel_path in self.bagged_size:
            # Already added, assume checksum OK
            return
        self.bagged_size[rel_path] = os.path.getsize(lpath)

        if SHA1 not in checksums:
            # ensure we always have sha1
            checksums = dict(checksums)
            with open(lpath, "rb") as file_path:
                # FIXME: Need sha-256 / sha-512 as well for Research Object BagIt profile?
                checksums[SHA1] = checksum_copy(file_path, hasher=hashlib.sha1)

        self.add_to_manifest(rel_path, checksums)

    def _relativise_files(
        self,
        structure: CWLObjectType | CWLOutputType | MutableSequence[CWLObjectType],
    ) -> None:
        """Save any file objects into the RO and update the local paths."""
        # Base case - we found a File we need to update
        _logger.debug("[provenance] Relativising: %s", structure)

        if isinstance(structure, MutableMapping):
            if structure.get("class") == "File":
                relative_path: str | PurePosixPath | None = None
                if "checksum" in structure:
                    raw_checksum = cast(str, structure["checksum"])
                    alg, checksum = raw_checksum.split("$")
                    if alg != SHA1:
                        raise TypeError(
                            f"Only SHA1 CWL checksums are currently supported: {structure}"
                        )
                    if self.has_data_file(checksum):
                        prefix = checksum[0:2]
                        relative_path = PurePosixPath("data") / prefix / checksum

                if not (relative_path is not None and "location" in structure):
                    # Register in RO; but why was this not picked
                    # up by used_artefacts?
                    _logger.info("[provenance] Adding to RO %s", structure["location"])
                    with self.fsaccess.open(cast(str, structure["location"]), "rb") as fp:
                        relative_path = self.add_data_file(fp)
                        checksum = PurePosixPath(relative_path).name
                        structure["checksum"] = f"{SHA1}${checksum}"
                # RO-relative path as new location
                structure["location"] = str(PurePosixPath("..") / relative_path)
                if "path" in structure:
                    del structure["path"]

            if structure.get("class") == "Directory":
                # TODO: Generate anonymous Directory with a "listing"
                # pointing to the hashed files
                del structure["location"]

            for val in structure.values():
                try:
                    self._relativise_files(val)
                except OSError:
                    pass
            return

        if isinstance(structure, MutableSequence):
            for obj in structure:
                # Recurse and rewrite any nested File objects
                self._relativise_files(obj)



import re
from pathlib import Path

import pytest

from .util import get_data, get_main_output, needs_docker

test_factors = [(""), ("--parallel"), ("--debug"), ("--parallel --debug")]


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_wf_without_container(tmp_path: Path, factor: str) -> None:
    """Confirm that we can run a workflow without a container."""
    test_file = "hello-workflow.cwl"
    cache_dir = str(tmp_path / "cwltool_cache")
    commands = factor.split()
    commands.extend(
        [
            "--cachedir",
            cache_dir,
            "--outdir",
            str(tmp_path / "outdir"),
            get_data("tests/wf/" + test_file),
            "--usermessage",
            "hello",
        ]
    )
    error_code, _, stderr = get_main_output(commands)

    stderr = re.sub(r"\s\s+", " ", stderr)
    assert "completed success" in stderr
    assert error_code == 0


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_issue_740_fixed(tmp_path: Path, factor: str) -> None:
    """Confirm that re-running a particular workflow with caching succeeds."""
    test_file = "cache_test_workflow.cwl"
    cache_dir = str(tmp_path / "cwltool_cache")
    commands = factor.split()
    commands.extend(["--cachedir", cache_dir, get_data("tests/wf/" + test_file)])
    error_code, _, stderr = get_main_output(commands)

    stderr = re.sub(r"\s\s+", " ", stderr)
    assert "completed success" in stderr
    assert error_code == 0

    commands = factor.split()
    commands.extend(["--cachedir", cache_dir, get_data("tests/wf/" + test_file)])
    error_code, _, stderr = get_main_output(commands)

    stderr = re.sub(r"\s\s+", " ", stderr)
    assert "Output of job will be cached in" not in stderr
    assert error_code == 0, stderr


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_cache_relative_paths(tmp_path: Path, factor: str) -> None:
    """Confirm that re-running a particular workflow with caching succeeds."""
    test_file = "secondary-files.cwl"
    test_job_file = "secondary-files-job.yml"
    cache_dir = str(tmp_path / "cwltool_cache")
    commands = factor.split()
    commands.extend(
        [
            "--out",
            str(tmp_path / "out"),
            "--cachedir",
            cache_dir,
            get_data(f"tests/{test_file}"),
            get_data(f"tests/{test_job_file}"),
        ]
    )
    error_code, _, stderr = get_main_output(commands)

    stderr = re.sub(r"\s\s+", " ", stderr)
    assert "completed success" in stderr
    assert error_code == 0

    commands = factor.split()
    commands.extend(
        [
            "--out",
            str(tmp_path / "out2"),
            "--cachedir",
            cache_dir,
            get_data(f"tests/{test_file}"),
            get_data(f"tests/{test_job_file}"),
        ]
    )
    error_code, _, stderr = get_main_output(commands)

    stderr = re.sub(r"\s\s+", " ", stderr)
    assert "Output of job will be cached in" not in stderr
    assert error_code == 0, stderr

    assert (tmp_path / "cwltool_cache" / "27903451fc1ee10c148a0bdeb845b2cf").exists()


@pytest.mark.parametrize("factor", test_factors)
def test_cache_default_literal_file(tmp_path: Path, factor: str) -> None:
    """Confirm that running a CLT with a default literal file with caching succeeds."""
    test_file = "tests/wf/extract_region_specs.cwl"
    cache_dir = str(tmp_path / "cwltool_cache")
    commands = factor.split()
    commands.extend(
        [
            "--out",
            str(tmp_path / "out"),
            "--cachedir",
            cache_dir,
            get_data(test_file),
        ]
    )
    error_code, _, stderr = get_main_output(commands)

    stderr = re.sub(r"\s\s+", " ", stderr)
    assert "completed success" in stderr
    assert error_code == 0


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_cache_dockerreq_hint_instead_of_req(tmp_path: Path, factor: str) -> None:
    """The cache must not be checked when there is an invalid use of an absolute path in iwdr.listing."""
    cache_dir = str(tmp_path / "cwltool_cache")
    test_job_file = "tests/wf/loadContents-input.yml"
    # First, run the iwd-container-entryname1 conformance tests with caching turned on
    test1_file = "tests/wf/iwd-container-entryname1.cwl"
    commands1 = factor.split()
    commands1.extend(
        [
            "--out",
            str(tmp_path / "out1"),
            "--cachedir",
            cache_dir,
            get_data(test1_file),
            get_data(test_job_file),
        ]
    )
    error_code1, _, stderr1 = get_main_output(commands1)

    stderr1 = re.sub(r"\s\s+", " ", stderr1)
    assert "completed success" in stderr1
    assert error_code1 == 0
    # Second, run the iwd-container-entryname3 test, which should fail
    # even though it would be a cache hit, except that its DockerRequirement is
    # in `hints` instead of `requirements` and one of the initial working directory
    # items has an absolute path starting with `/`.
    test2_file = "tests/wf/iwd-container-entryname3.cwl"
    commands2 = factor.split()
    commands2.extend(
        [
            "--out",
            str(tmp_path / "out2"),
            "--cachedir",
            cache_dir,
            get_data(test2_file),
            get_data(test_job_file),
        ]
    )
    error_code2, _, stderr2 = get_main_output(commands2)

    stderr2 = re.sub(r"\s\s+", " ", stderr2)
    assert (
        "at index 0 of listing is invalid, name can only start with '/' "
        "when DockerRequirement is in 'requirements" in stderr2
    )
    assert error_code2 == 1



"""Single and multi-threaded executors."""

import datetime
import functools
import logging
import math
import os
import sys
import threading
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable, MutableSequence
from threading import Lock
from typing import Optional, cast

import psutil
from mypy_extensions import mypyc_attr
from schema_salad.exceptions import ValidationException
from schema_salad.sourceline import SourceLine

from .command_line_tool import CallbackJob, ExpressionJob
from .context import RuntimeContext, getdefault
from .cuda import cuda_version_and_device_count
from .errors import WorkflowException
from .job import JobBase
from .loghandler import _logger
from .mutation import MutationManager
from .process import Process, cleanIntermediate, relocateOutputs
from .task_queue import TaskQueue
from .update import ORIGINAL_CWLVERSION
from .utils import CWLObjectType, JobsType
from .workflow import Workflow
from .workflow_job import WorkflowJob, WorkflowJobStep

TMPDIR_LOCK = Lock()


@mypyc_attr(allow_interpreted_subclasses=True)
class JobExecutor(metaclass=ABCMeta):
    """Abstract base job executor."""

    def __init__(self) -> None:
        """Initialize."""
        self.final_output: MutableSequence[CWLObjectType | None] = []
        self.final_status: list[str] = []
        self.output_dirs: set[str] = set()

    def __call__(
        self,
        process: Process,
        job_order_object: CWLObjectType,
        runtime_context: RuntimeContext,
        logger: logging.Logger = _logger,
    ) -> tuple[CWLObjectType | None, str]:
        return self.execute(process, job_order_object, runtime_context, logger)

    def output_callback(self, out: CWLObjectType | None, process_status: str) -> None:
        """Collect the final status and outputs."""
        self.final_status.append(process_status)
        self.final_output.append(out)

    @abstractmethod
    def run_jobs(
        self,
        process: Process,
        job_order_object: CWLObjectType,
        logger: logging.Logger,
        runtime_context: RuntimeContext,
    ) -> None:
        """Execute the jobs for the given Process."""

    def execute(
        self,
        process: Process,
        job_order_object: CWLObjectType,
        runtime_context: RuntimeContext,
        logger: logging.Logger = _logger,
    ) -> tuple[CWLObjectType | None, str]:
        """Execute the process."""

        self.final_output = []
        self.final_status = []

        if not runtime_context.basedir:
            raise WorkflowException("Must provide 'basedir' in runtimeContext")

        def check_for_abstract_op(tool: CWLObjectType) -> None:
            if tool["class"] == "Operation":
                raise SourceLine(tool, "class", WorkflowException, runtime_context.debug).makeError(
                    "Workflow has unrunnable abstract Operation"
                )

        process.visit(check_for_abstract_op)

        finaloutdir = None  # Type: Optional[str]
        original_outdir = runtime_context.outdir
        if isinstance(original_outdir, str):
            finaloutdir = os.path.abspath(original_outdir)
        runtime_context = runtime_context.copy()
        outdir = runtime_context.create_outdir()
        self.output_dirs.add(outdir)
        runtime_context.outdir = outdir
        runtime_context.mutation_manager = MutationManager()
        runtime_context.toplevel = True
        runtime_context.workflow_eval_lock = threading.Condition(threading.RLock())

        job_reqs: list[CWLObjectType] | None = None
        if "https://w3id.org/cwl/cwl#requirements" in job_order_object:
            if process.metadata.get(ORIGINAL_CWLVERSION) == "v1.0":
                raise WorkflowException(
                    "`cwl:requirements` in the input object is not part of CWL "
                    "v1.0. You can adjust to use `cwltool:overrides` instead; or you "
                    "can set the cwlVersion to v1.1"
                )
            job_reqs = cast(
                list[CWLObjectType],
                job_order_object["https://w3id.org/cwl/cwl#requirements"],
            )
        elif "cwl:defaults" in process.metadata and "https://w3id.org/cwl/cwl#requirements" in cast(
            CWLObjectType, process.metadata["cwl:defaults"]
        ):
            if process.metadata.get(ORIGINAL_CWLVERSION) == "v1.0":
                raise WorkflowException(
                    "`cwl:requirements` in the input object is not part of CWL "
                    "v1.0. You can adjust to use `cwltool:overrides` instead; or you "
                    "can set the cwlVersion to v1.1"
                )
            job_reqs = cast(
                Optional[list[CWLObjectType]],
                cast(CWLObjectType, process.metadata["cwl:defaults"])[
                    "https://w3id.org/cwl/cwl#requirements"
                ],
            )
        if job_reqs is not None:
            for req in job_reqs:
                process.requirements.append(req)

        self.run_jobs(process, job_order_object, logger, runtime_context)
        if runtime_context.validate_only is True:
            return (None, "ValidationSuccess")

        if self.final_output and self.final_output[0] is not None and finaloutdir is not None:
            self.final_output[0] = relocateOutputs(
                self.final_output[0],
                finaloutdir,
                self.output_dirs,
                runtime_context.move_outputs,
                runtime_context.make_fs_access(""),
                getdefault(runtime_context.compute_checksum, True),
                path_mapper=runtime_context.path_mapper,
            )

        if runtime_context.rm_tmpdir:
            if not runtime_context.cachedir:
                output_dirs: Iterable[str] = self.output_dirs
            else:
                output_dirs = filter(
                    lambda x: not x.startswith(runtime_context.cachedir),  # type: ignore
                    self.output_dirs,
                )
            cleanIntermediate(output_dirs)

        if self.final_output and self.final_status:
            if (
                runtime_context.research_obj is not None
                and isinstance(process, (JobBase, Process, WorkflowJobStep, WorkflowJob))
                and process.parent_wf
            ):
                process_run_id: str | None = None
                name = "primary"
                process.parent_wf.generate_output_prov(self.final_output[0], process_run_id, name)
                process.parent_wf.document.wasEndedBy(
                    process.parent_wf.workflow_run_uri,
                    None,
                    process.parent_wf.engine_uuid,
                    datetime.datetime.now(),
                )
                process.parent_wf.finalize_prov_profile(name=None)
            return (self.final_output[0], self.final_status[0])
        return (None, "permanentFail")


@mypyc_attr(allow_interpreted_subclasses=True)
class SingleJobExecutor(JobExecutor):
    """Default single-threaded CWL reference executor."""

    def run_jobs(
        self,
        process: Process,
        job_order_object: CWLObjectType,
        logger: logging.Logger,
        runtime_context: RuntimeContext,
    ) -> None:
        process_run_id: str | None = None

        # define provenance profile for single commandline tool
        if not isinstance(process, Workflow) and runtime_context.research_obj is not None:
            process.provenance_object = runtime_context.research_obj.initialize_provenance(
                full_name=runtime_context.cwl_full_name,
                # following are only set from main when directly command line tool
                # when nested in a workflow, they should be disabled since they would
                # already have been provided/initialized by the parent workflow prov-obj
                host_provenance=runtime_context.prov_host,
                user_provenance=runtime_context.prov_user,
                orcid=runtime_context.orcid,
                # single tool execution, so RO UUID = wf UUID = tool UUID
                run_uuid=runtime_context.research_obj.ro_uuid,
                fsaccess=runtime_context.make_fs_access(""),
            )
            process.parent_wf = process.provenance_object

        jobiter = process.job(job_order_object, self.output_callback, runtime_context)

        try:
            for job in jobiter:
                if job is not None:
                    if runtime_context.builder is not None and hasattr(job, "builder"):
                        job.builder = runtime_context.builder
                    if job.outdir is not None:
                        self.output_dirs.add(job.outdir)
                    if runtime_context.research_obj is not None:
                        if not isinstance(process, Workflow):
                            prov_obj = process.provenance_object
                        else:
                            prov_obj = job.prov_obj
                        if prov_obj:
                            runtime_context.prov_obj = prov_obj
                            prov_obj.fsaccess = runtime_context.make_fs_access("")
                            prov_obj.evaluate(
                                process,
                                job,
                                job_order_object,
                                runtime_context.research_obj,
                            )
                            process_run_id = prov_obj.record_process_start(process, job)
                            runtime_context = runtime_context.copy()
                        runtime_context.process_run_id = process_run_id
                    if runtime_context.validate_only is True:
                        if isinstance(job, WorkflowJob):
                            name = job.tool.lc.filename
                        else:
                            name = getattr(job, "name", str(job))
                        print(
                            f"{name} is valid CWL. No errors detected in the inputs.",
                            file=runtime_context.validate_stdout,
                        )
                        return
                    job.run(runtime_context)
                else:
                    logger.error("Workflow cannot make any more progress.")
                    break
        except (
            ValidationException,
            WorkflowException,
        ):  # pylint: disable=try-except-raise
            raise
        except Exception as err:
            logger.exception("Got workflow error")
            raise WorkflowException(str(err)) from err


if sys.platform != "darwin":

    def _max_cores() -> int:
        return len(psutil.Process().cpu_affinity())

else:

    def _max_cores() -> int:
        return psutil.cpu_count() or 1


class MultithreadedJobExecutor(JobExecutor):
    """
    Experimental multi-threaded CWL executor.

    Does simple resource accounting, will not start a job unless it
    has cores / ram available, but does not make any attempt to
    optimize usage.
    """

    def __init__(self, max_parallel: int = 0) -> None:
        """Initialize."""
        super().__init__()
        self.exceptions: list[WorkflowException] = []
        self.pending_jobs: list[JobsType] = []
        self.pending_jobs_lock = threading.Lock()

        self.max_ram = int(psutil.virtual_memory().available / 2**20)
        self.max_cores = float(max_parallel if max_parallel >= 1 else _max_cores())
        self.max_cuda = cuda_version_and_device_count()[1]
        self.allocated_ram = float(0)
        self.allocated_cores = float(0)
        self.allocated_cuda: int = 0

    def select_resources(
        self, request: dict[str, int | float], runtime_context: RuntimeContext
    ) -> dict[str, int | float]:  # pylint: disable=unused-argument
        """Naïve check for available cpu cores and memory."""
        result: dict[str, int | float] = {}
        maxrsc = {"cores": self.max_cores, "ram": self.max_ram}
        resources_types = {"cores", "ram"}
        if "cudaDeviceCountMin" in request or "cudaDeviceCountMax" in request:
            maxrsc["cudaDeviceCount"] = self.max_cuda
            resources_types.add("cudaDeviceCount")
        for rsc in resources_types:
            rsc_min = request[rsc + "Min"]
            if rsc_min > maxrsc[rsc]:
                raise WorkflowException(
                    f"Requested at least {rsc_min} {rsc} but only " f"{maxrsc[rsc]} available"
                )
            rsc_max = request[rsc + "Max"]
            if rsc_max < maxrsc[rsc]:
                result[rsc] = math.ceil(rsc_max)
            else:
                result[rsc] = maxrsc[rsc]

        result["tmpdirSize"] = math.ceil(request["tmpdirMin"])
        result["outdirSize"] = math.ceil(request["outdirMin"])

        return result

    def _runner(
        self,
        job: JobBase | WorkflowJob | CallbackJob | ExpressionJob,
        runtime_context: RuntimeContext,
        TMPDIR_LOCK: threading.Lock,
    ) -> None:
        """Job running thread."""
        try:
            _logger.debug(
                "job: {}, runtime_context: {}, TMPDIR_LOCK: {}".format(
                    job, runtime_context, TMPDIR_LOCK
                )
            )
            job.run(runtime_context, TMPDIR_LOCK)
        except WorkflowException as err:
            _logger.exception(f"Got workflow error: {err}")
            self.exceptions.append(err)
        except Exception as err:  # pylint: disable=broad-except
            _logger.exception(f"Got workflow error: {err}")
            wf_exc = WorkflowException(str(err))
            wf_exc.__cause__ = err
            wf_exc.__suppress_context__ = True
            self.exceptions.append(wf_exc)
        finally:
            if runtime_context.workflow_eval_lock:
                with runtime_context.workflow_eval_lock:
                    if isinstance(job, JobBase):
                        ram = job.builder.resources["ram"]
                        self.allocated_ram -= ram
                        cores = job.builder.resources["cores"]
                        self.allocated_cores -= cores
                        cudaDevices: int = cast(
                            int, job.builder.resources.get("cudaDeviceCount", 0)
                        )
                        self.allocated_cuda -= cudaDevices
                    runtime_context.workflow_eval_lock.notify_all()

    def run_job(
        self,
        job: JobsType | None,
        runtime_context: RuntimeContext,
    ) -> None:
        """Execute a single Job in a separate thread."""
        if job is not None:
            with self.pending_jobs_lock:
                self.pending_jobs.append(job)

        with self.pending_jobs_lock:
            n = 0
            while (n + 1) <= len(self.pending_jobs):
                # Simple greedy resource allocation strategy.  Go
                # through pending jobs in the order they were
                # generated and add them to the queue only if there
                # are resources available.
                job = self.pending_jobs[n]
                if isinstance(job, JobBase):
                    ram = job.builder.resources["ram"]
                    cores = job.builder.resources["cores"]
                    cudaDevices = cast(int, job.builder.resources.get("cudaDeviceCount", 0))
                    if ram > self.max_ram or cores > self.max_cores or cudaDevices > self.max_cuda:
                        _logger.error(
                            'Job "%s" cannot be run, requests more resources (%s) '
                            "than available on this host (already allocated ram is %d, "
                            "allocated cores is %d, allocated CUDA is %d, "
                            "max ram %d, max cores %d, max CUDA %d).",
                            job.name,
                            job.builder.resources,
                            self.allocated_ram,
                            self.allocated_cores,
                            self.allocated_cuda,
                            self.max_ram,
                            self.max_cores,
                            self.max_cuda,
                        )
                        self.pending_jobs.remove(job)
                        return

                    if (
                        self.allocated_ram + ram > self.max_ram
                        or self.allocated_cores + cores > self.max_cores
                        or self.allocated_cuda + cudaDevices > self.max_cuda
                    ):
                        _logger.debug(
                            'Job "%s" cannot run yet, resources (%s) are not '
                            "available (already allocated ram is %d, allocated cores is %d, "
                            "allocated CUDA devices is %d, "
                            "max ram %d, max cores %d, max CUDA %d).",
                            job.name,
                            job.builder.resources,
                            self.allocated_ram,
                            self.allocated_cores,
                            self.allocated_cuda,
                            self.max_ram,
                            self.max_cores,
                            self.max_cuda,
                        )
                        n += 1
                        continue

                if isinstance(job, JobBase):
                    ram = job.builder.resources["ram"]
                    self.allocated_ram += ram
                    cores = job.builder.resources["cores"]
                    self.allocated_cores += cores
                    cuda = cast(int, job.builder.resources.get("cudaDevices", 0))
                    self.allocated_cuda += cuda
                self.taskqueue.add(
                    functools.partial(self._runner, job, runtime_context, TMPDIR_LOCK),
                    runtime_context.workflow_eval_lock,
                )
                self.pending_jobs.remove(job)

    def wait_for_next_completion(self, runtime_context: RuntimeContext) -> None:
        """Wait for jobs to finish."""
        if runtime_context.workflow_eval_lock is not None:
            runtime_context.workflow_eval_lock.wait(timeout=3)
        if self.exceptions:
            raise self.exceptions[0]

    def run_jobs(
        self,
        process: Process,
        job_order_object: CWLObjectType,
        logger: logging.Logger,
        runtime_context: RuntimeContext,
    ) -> None:
        self.taskqueue: TaskQueue = TaskQueue(threading.Lock(), int(math.ceil(self.max_cores)))
        try:
            jobiter = process.job(job_order_object, self.output_callback, runtime_context)

            if runtime_context.workflow_eval_lock is None:
                raise WorkflowException("runtimeContext.workflow_eval_lock must not be None")

            runtime_context.workflow_eval_lock.acquire()
            for job in jobiter:
                if job is not None:
                    if isinstance(job, JobBase):
                        job.builder = runtime_context.builder or job.builder
                        if job.outdir is not None:
                            self.output_dirs.add(job.outdir)

                self.run_job(job, runtime_context)

                if job is None:
                    if self.taskqueue.in_flight > 0:
                        self.wait_for_next_completion(runtime_context)
                    else:
                        logger.error("Workflow cannot make any more progress.")
                        break

            self.run_job(None, runtime_context)
            while self.taskqueue.in_flight > 0:
                self.wait_for_next_completion(runtime_context)
                self.run_job(None, runtime_context)

            runtime_context.workflow_eval_lock.release()
        finally:
            self.taskqueue.drain()
            self.taskqueue.join()


class NoopJobExecutor(JobExecutor):
    """Do nothing executor, for testing purposes only."""

    def run_jobs(
        self,
        process: Process,
        job_order_object: CWLObjectType,
        logger: logging.Logger,
        runtime_context: RuntimeContext,
    ) -> None:
        pass

    def execute(
        self,
        process: Process,
        job_order_object: CWLObjectType,
        runtime_context: RuntimeContext,
        logger: logging.Logger | None = None,
    ) -> tuple[CWLObjectType | None, str]:
        return {}, "success"



"""
Confirm that we can subclass and/or serializae certain classes used by 3rd parties.

Especially if those classes are (or become) compiled with mypyc.
"""

import pickle

import pytest
from ruamel.yaml.comments import CommentedMap
from schema_salad.avro import schema

from cwltool.builder import Builder
from cwltool.command_line_tool import CommandLineTool, ExpressionTool
from cwltool.context import LoadingContext, RuntimeContext
from cwltool.procgenerator import ProcessGenerator
from cwltool.stdfsaccess import StdFsAccess
from cwltool.update import INTERNAL_VERSION
from cwltool.workflow import Workflow

from .test_anon_types import snippet
from .util import get_data


@pytest.mark.parametrize("snippet", snippet)
def test_subclass_CLT(snippet: CommentedMap) -> None:
    """We can subclass CommandLineTool."""

    class TestCLT(CommandLineTool):
        test = True

    a = TestCLT(snippet, LoadingContext())
    assert a.test is True


@pytest.mark.parametrize("snippet", snippet)
def test_subclass_exprtool(snippet: CommentedMap) -> None:
    """We can subclass ExpressionTool."""

    class TestExprTool(ExpressionTool):
        test = False

    a = TestExprTool(snippet, LoadingContext())
    assert a.test is False


@pytest.mark.parametrize("snippet", snippet)
def test_pickle_unpickle_workflow(snippet: CommentedMap) -> None:
    """We can pickle & unpickle a Workflow."""

    a = Workflow(snippet, LoadingContext())
    stream = pickle.dumps(a)
    assert stream
    assert pickle.loads(stream)


def test_pickle_processgenerator() -> None:
    """We can pickle ProcessGenerator."""

    pytoolgen = get_data("tests/wf/generator/pytoolgen.cwl")
    a = ProcessGenerator(
        CommentedMap(
            {
                "cwlVersion": "v1.0",
                "id": f"file://{pytoolgen}",
                "$namespaces": {"cwltool": "http://commonwl.org/cwltool#"},
                "class": "cwltool:ProcessGenerator",
                "inputs": {},
                "outputs": {},
                "run": {
                    "id": f"file://{pytoolgen}#f9f617af-3088-4ade-bbf3-acd1d6f51355",
                    "cwlVersion": "v1.0",
                    "class": "CommandLineTool",
                    "inputs": {},
                    "outputs": {},
                    "baseCommand": "echo",
                },
            },
        ),
        LoadingContext(),
    )

    assert pickle.dumps(a)


def test_serialize_builder() -> None:
    """We can pickle Builder."""
    runtime_context = RuntimeContext()
    builder = Builder(
        {},
        [],
        [],
        {},
        schema.Names(),
        [],
        [],
        {},
        None,
        None,
        StdFsAccess,
        StdFsAccess(""),
        None,
        0.1,
        False,
        False,
        False,
        "no_listing",
        runtime_context.get_outdir(),
        runtime_context.get_tmpdir(),
        runtime_context.get_stagedir(),
        INTERNAL_VERSION,
        "docker",
    )
    assert pickle.dumps(builder)


def test_pickle_unpickle_runtime_context() -> None:
    """We can pickle & unpickle RuntimeContext"""

    runtime_context = RuntimeContext()
    stream = pickle.dumps(runtime_context)
    assert stream
    assert pickle.loads(stream)
